from io import BytesIO
import os
import urllib

import pandas as pd

from aif360.sklearn.datasets.utils import standardize_dataset


# cache location
DATA_HOME_DEFAULT = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 '..', 'data', 'raw')
MEPS_URL = "https://meps.ahrq.gov/mepsweb/data_files/pufs"
PROMPT = """
By using this function you acknowledge the responsibility for reading and
abiding by any copyright/usage rules and restrictions as stated on the MEPS web
site (https://meps.ahrq.gov/data_stats/data_use.jsp).

Continue [y/n]? > """

def fetch_meps(panel, *, accept_terms=None, data_home=None, cache=True,
               usecols=['REGION', 'AGE', 'SEX', 'RACE', 'MARRY', 'FTSTU',
                        'ACTDTY', 'HONRDC', 'RTHLTH', 'MNHLTH', 'HIBPDX',
                        'CHDDX', 'ANGIDX', 'MIDX', 'OHRTDX', 'STRKDX', 'EMPHDX',
                        'CHBRON', 'CHOLDX', 'CANCERDX', 'DIABDX', 'JTPAIN',
                        'ARTHDX', 'ARTHTYPE', 'ASTHDX', 'ADHDADDX', 'PREGNT',
                        'WLKLIM', 'ACTLIM', 'SOCLIM', 'COGLIM', 'DFHEAR42',
                        'DFSEE42', 'ADSMOK42', 'PCS42', 'MCS42', 'K6SUM42',
                        'PHQ242', 'EMPST', 'POVCAT', 'INSCOV'],
               dropcols=None, numeric_only=False, dropna=True):
    """Load the Medical Expenditure Panel Survey (MEPS) dataset.

    Note:
        For descriptions of the dataset features, see the `data codebook
        <https://meps.ahrq.gov/mepsweb/data_stats/download_data_files_codebook.jsp?PUFId=H181>`_.

    Args:
        panel ({19, 20, 21}): Panel number (only 19, 20, and 21 are currently
            supported).
        accept_terms (bool, optional): Bypass terms prompt. Note: by setting
            this to ``True``, you acknowledge responsibility for reading and
            accepting the MEPS usage terms.
        data_home (string, optional): Specify another download and cache folder
            for the datasets. By default all AIF360 datasets are stored in
            'aif360/sklearn/data/raw' subfolders.
        cache (bool): Whether to cache downloaded datasets.
        usecols (single label or list-like, optional): Feature column(s) to
            keep. All others are dropped.
        dropcols (single label or list-like, optional): Feature column(s) to
            drop.
        numeric_only (bool): Drop all non-numeric feature columns.
        dropna (bool): Drop rows with NAs.

    Returns:
        namedtuple: Tuple containing X and y for the MEPS dataset accessible by
        index or name.
    """
    if panel not in {19, 20, 21}:
        raise ValueError("only panels 19, 20, and 21 are currently supported.")

    fname = 'h192ssp.zip' if panel == 21 else 'h181ssp.zip'
    cache_path = os.path.join(data_home or DATA_HOME_DEFAULT, fname)
    if cache and os.path.isfile(cache_path):
        df = pd.read_sas(cache_path, format="xport", encoding="utf-8")
    else:
        # skip prompt if user chooses
        accept = accept_terms or input(PROMPT)
        if accept != 'y' and accept is not True:
            raise PermissionError("Terms not agreed.")
        rawz = urllib.request.urlopen(os.path.join(MEPS_URL, fname)).read()
        df = pd.read_sas(BytesIO(rawz), format='xport', encoding="utf-8", compression="zip")
        if cache:
            os.makedirs(os.path.dirname(cache_path), exist_ok=True)
            with open(cache_path, "wb") as f:
                f.write(rawz)
    # restrict to correct panel
    df = df[df['PANEL'] == panel]
    # change all 15s to 16s if panel == 21
    yr = 16 if panel == 21 else 15

    # non-Hispanic Whites are marked as WHITE; all others as NON-WHITE
    df['RACEV2X'] = (df['HISPANX'] == 2) & (df['RACEV2X'] == 1)

    # rename all columns that are panel/round-specific
    df = df.rename(columns={
        'FTSTU53X': 'FTSTU', 'ACTDTY53': 'ACTDTY', 'HONRDC53': 'HONRDC',
        'RTHLTH53': 'RTHLTH', 'MNHLTH53': 'MNHLTH', 'CHBRON53': 'CHBRON',
        'JTPAIN53': 'JTPAIN', 'PREGNT53': 'PREGNT', 'WLKLIM53': 'WLKLIM',
        'ACTLIM53': 'ACTLIM', 'SOCLIM53': 'SOCLIM', 'COGLIM53': 'COGLIM',
        'EMPST53': 'EMPST', 'REGION53': 'REGION', 'MARRY53X': 'MARRY',
        'AGE53X': 'AGE', f'POVCAT{yr}': 'POVCAT', f'INSCOV{yr}': 'INSCOV',
        f'PERWT{yr}F': 'PERWT', 'RACEV2X': 'RACE'})

    df.loc[df.AGE < 0, 'AGE'] = None  # set invalid ages to NaN
    cat_cols = ['REGION', 'SEX', 'RACE', 'MARRY', 'FTSTU', 'ACTDTY', 'HONRDC',
                'RTHLTH', 'MNHLTH', 'HIBPDX', 'CHDDX', 'ANGIDX', 'MIDX',
                'OHRTDX', 'STRKDX', 'EMPHDX', 'CHBRON', 'CHOLDX', 'CANCERDX',
                'DIABDX', 'JTPAIN', 'ARTHDX', 'ARTHTYPE', 'ASTHDX', 'ADHDADDX',
                'PREGNT', 'WLKLIM', 'ACTLIM', 'SOCLIM', 'COGLIM', 'DFHEAR42',
                'DFSEE42', 'ADSMOK42', 'PHQ242', 'EMPST', 'POVCAT', 'INSCOV',
    # NOTE: education tracking seems to have changed between panels. 'EDUYRDG'
    # was used for panel 19, 'EDUCYR' and 'HIDEG' were used for panels 20 & 21.
    # User may change usecols to include these manually.
                'EDUCYR', 'HIDEG']
    if panel == 19:
        cat_cols += ['EDUYRDG']

    for col in cat_cols:
        df[col] = df[col].astype('category')
        thresh = 0 if col in ['REGION', 'MARRY', 'ASTHDX'] else -1
        na_cats = [c for c in df[col].cat.categories if c < thresh]
        df[col] = df[col].cat.remove_categories(na_cats)  # set NaN cols to NaN

    df['SEX'] = df['SEX'].cat.rename_categories({1: 'Male', 2: 'Female'})
    df['RACE'] = df['RACE'].cat.rename_categories({False: 'Non-White', True: 'White'})
    df['RACE'] = df['RACE'].cat.reorder_categories(['Non-White', 'White'], ordered=True)

    # Compute UTILIZATION, binarize it to 0 (< 10) and 1 (>= 10)
    cols = [f'OBTOTV{yr}', f'OPTOTV{yr}', f'ERTOT{yr}', f'IPNGTD{yr}', f'HHTOTD{yr}']
    util = df[cols].sum(axis=1)
    df['UTILIZATION'] = pd.cut(util, [min(util)-1, 10, max(util)+1], right=False,
                               labels=['< 10 Visits', '>= 10 Visits'])#['low', 'high'])

    return standardize_dataset(df, prot_attr='RACE', target='UTILIZATION',
                               sample_weight='PERWT', usecols=usecols,
                               dropcols=dropcols, numeric_only=numeric_only,
                               dropna=dropna)



from aif360.detectors.mdss.ScoringFunctions.ScoringFunction import ScoringFunction
from aif360.detectors.mdss.ScoringFunctions.Bernoulli import Bernoulli
from aif360.detectors.mdss.ScoringFunctions.Poisson import Poisson
from aif360.detectors.mdss.ScoringFunctions.BerkJones import BerkJones
from aif360.detectors.mdss.ScoringFunctions.Gaussian import Gaussian



import numpy as np

from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC as SVM
from sklearn.preprocessing import MinMaxScaler

from aif360.algorithms.preprocessing import DisparateImpactRemover
from aif360.datasets import AdultDataset
from aif360.metrics import ClassificationMetric


def test_repair0():
    ad = AdultDataset(protected_attribute_names=['sex'],
        privileged_classes=[['Male']], categorical_features=[],
        features_to_keep=['age', 'education-num'])

    di = DisparateImpactRemover(repair_level=0.)
    ad_repd = di.fit_transform(ad)

    assert ad_repd == ad

def test_adult():
    protected = 'sex'
    ad = AdultDataset(protected_attribute_names=[protected],
        privileged_classes=[['Male']], categorical_features=[],
        features_to_keep=['age', 'education-num', 'capital-gain', 'capital-loss', 'hours-per-week'])

    scaler = MinMaxScaler(copy=False)
    # ad.features = scaler.fit_transform(ad.features)

    test, train = ad.split([16281])
    assert np.any(test.labels)

    train.features = scaler.fit_transform(train.features)
    test.features = scaler.transform(test.features)

    index = train.feature_names.index(protected)
    X_tr = np.delete(train.features, index, axis=1)
    X_te = np.delete(test.features, index, axis=1)
    y_tr = train.labels.ravel()

    di = DisparateImpactRemover(repair_level=1.0)
    train_repd = di.fit_transform(train)
    # train_repd2 = di.fit_transform(train)
    # assert train_repd == train_repd2
    test_repd = di.fit_transform(test)

    assert np.all(train_repd.protected_attributes == train.protected_attributes)

    lmod = LogisticRegression(class_weight='balanced')
    # lmod = SVM(class_weight='balanced')
    lmod.fit(X_tr, y_tr)

    test_pred = test.copy()
    test_pred.labels = lmod.predict(X_te)

    X_tr_repd = np.delete(train_repd.features, index, axis=1)
    X_te_repd = np.delete(test_repd.features, index, axis=1)
    y_tr_repd = train_repd.labels.ravel()
    assert (y_tr == y_tr_repd).all()

    lmod.fit(X_tr_repd, y_tr_repd)
    test_repd_pred = test_repd.copy()
    test_repd_pred.labels = lmod.predict(X_te_repd)

    p = [{protected: 1}]
    u = [{protected: 0}]

    cm = ClassificationMetric(test, test_pred, privileged_groups=p, unprivileged_groups=u)
    before = cm.disparate_impact()
    # print('Disparate impact: {:.4}'.format(before))
    # print('Acc overall: {:.4}'.format(cm.accuracy()))

    repaired_cm = ClassificationMetric(test_repd, test_repd_pred, privileged_groups=p, unprivileged_groups=u)
    after = repaired_cm.disparate_impact()
    # print('Disparate impact: {:.4}'.format(after))
    # print('Acc overall: {:.4}'.format(repaired_cm.accuracy()))

    assert after > before
    assert abs(1 - after) <= 0.2



import os

import pandas as pd

from aif360.sklearn.datasets.utils import standardize_dataset


# cache location
DATA_HOME_DEFAULT = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 '..', 'data', 'raw')
COMPAS_URL = 'https://raw.githubusercontent.com/propublica/compas-analysis/bafff5da3f2e45eca6c2d5055faad269defd135a/compas-scores-two-years.csv'
COMPAS_VIOLENT_URL = 'https://raw.githubusercontent.com/propublica/compas-analysis/bafff5da3f2e45eca6c2d5055faad269defd135a/compas-scores-two-years-violent.csv'

def fetch_compas(subset='all', *, data_home=None, cache=True, binary_race=False,
                 usecols=['sex', 'age', 'age_cat', 'race', 'juv_fel_count',
                          'juv_misd_count', 'juv_other_count', 'priors_count',
                          'c_charge_degree', 'c_charge_desc'],
                 dropcols=None, numeric_only=False, dropna=True):
    """Load the COMPAS Recidivism Risk Scores dataset.

    Optionally binarizes 'race' to 'Caucasian' (privileged) or
    'African-American' (unprivileged). The other protected attribute is 'sex'
    ('Male' is *unprivileged* and 'Female' is *privileged*). The outcome
    variable is 'Survived' (favorable) if the person was not accused of a crime
    within two years or 'Recidivated' (unfavorable) if they were.

    Note:
        The values for the 'sex' variable if numeric_only is ``True`` are 1 for
        'Female and 0 for 'Male' -- opposite the convention of other datasets.

    Args:
        subset ({'all' or 'violent'}): Use the violent recidivism or full
            version of the dataset. Note: 'violent' is not a strict subset of
            'all' -- there are four samples in 'violent' which do not show up in
            'all'.
        data_home (string, optional): Specify another download and cache folder
            for the datasets. By default all AIF360 datasets are stored in
            'aif360/sklearn/data/raw' subfolders.
        cache (bool): Whether to cache downloaded datasets.
        binary_race (bool, optional): Filter only White and Black defendants.
        usecols (single label or list-like, optional): Feature column(s) to
            keep. All others are dropped.
        dropcols (single label or list-like, optional): Feature column(s) to
            drop.
        numeric_only (bool): Drop all non-numeric feature columns.
        dropna (bool): Drop rows with NAs.

    Returns:
        namedtuple: Tuple containing X and y for the COMPAS dataset accessible
        by index or name.
    """
    if subset not in {'violent', 'all'}:
        raise ValueError("subset must be either 'violent' or 'all'; cannot be "
                        f"{subset}")

    data_url = COMPAS_VIOLENT_URL if subset == 'violent' else COMPAS_URL
    cache_path = os.path.join(data_home or DATA_HOME_DEFAULT,
                              os.path.basename(data_url))
    if cache and os.path.isfile(cache_path):
        df = pd.read_csv(cache_path, index_col='id')
    else:
        df = pd.read_csv(data_url, index_col='id')
        if cache:
            os.makedirs(os.path.dirname(cache_path), exist_ok=True)
            df.to_csv(cache_path)

    # Perform the same preprocessing as the original analysis:
    # https://github.com/propublica/compas-analysis/blob/master/Compas%20Analysis.ipynb
    df = df[(df.days_b_screening_arrest <= 30)
          & (df.days_b_screening_arrest >= -30)
          & (df.is_recid != -1)
          & (df.c_charge_degree != 'O')
          & (df['score_text' if subset == 'all' else 'v_score_text'] != 'N/A')]

    for col in ['sex', 'age_cat', 'race', 'c_charge_degree', 'c_charge_desc']:
        df[col] = df[col].astype('category')

    # Misdemeanor < Felony
    df.c_charge_degree = df.c_charge_degree.cat.reorder_categories(
        ['M', 'F'], ordered=True)
    # 'Less than 25' < '25 - 45' < 'Greater than 45'
    df.age_cat = df.age_cat.cat.reorder_categories(
        ['Less than 25', '25 - 45', 'Greater than 45'], ordered=True)

    # 'Survived' < 'Recidivated'
    cats = ['Survived', 'Recidivated']
    df.two_year_recid = df.two_year_recid.replace([0, 1], cats).astype('category')
    df.two_year_recid = df.two_year_recid.cat.set_categories(cats, ordered=True)

    if binary_race:
        # 'African-American' < 'Caucasian'
        df.race = df.race.cat.set_categories(['African-American', 'Caucasian'],
                                             ordered=True)

    # 'Male' < 'Female'
    df.sex = df.sex.astype('category').cat.reorder_categories(
            ['Male', 'Female'], ordered=True)

    return standardize_dataset(df, prot_attr=['sex', 'race'],
                               target='two_year_recid', usecols=usecols,
                               dropcols=dropcols, numeric_only=numeric_only,
                               dropna=dropna)



import os

import pandas as pd

from aif360.datasets import StandardDataset


default_mappings = {
    'label_maps': [{1.0: '>50K', 0.0: '<=50K'}],
    'protected_attribute_maps': [{1.0: 'White', 0.0: 'Non-white'},
                                 {1.0: 'Male', 0.0: 'Female'}]
}

class AdultDataset(StandardDataset):
    """Adult Census Income Dataset.

    See :file:`aif360/data/raw/adult/README.md`.
    """

    def __init__(self, label_name='income-per-year',
                 favorable_classes=['>50K', '>50K.'],
                 protected_attribute_names=['race', 'sex'],
                 privileged_classes=[['White'], ['Male']],
                 instance_weights_name=None,
                 categorical_features=['workclass', 'education',
                     'marital-status', 'occupation', 'relationship',
                     'native-country'],
                 features_to_keep=[], features_to_drop=['fnlwgt'],
                 na_values=['?'], custom_preprocessing=None,
                 metadata=default_mappings):
        """See :obj:`StandardDataset` for a description of the arguments.

        Examples:
            The following will instantiate a dataset which uses the `fnlwgt`
            feature:

            >>> from aif360.datasets import AdultDataset
            >>> ad = AdultDataset(instance_weights_name='fnlwgt',
            ... features_to_drop=[])
            WARNING:root:Missing Data: 3620 rows removed from dataset.
            >>> not np.all(ad.instance_weights == 1.)
            True

            To instantiate a dataset which utilizes only numerical features and
            a single protected attribute, run:

            >>> single_protected = ['sex']
            >>> single_privileged = [['Male']]
            >>> ad = AdultDataset(protected_attribute_names=single_protected,
            ... privileged_classes=single_privileged,
            ... categorical_features=[],
            ... features_to_keep=['age', 'education-num'])
            >>> print(ad.feature_names)
            ['education-num', 'age', 'sex']
            >>> print(ad.label_names)
            ['income-per-year']

            Note: the `protected_attribute_names` and `label_name` are kept even
            if they are not explicitly given in `features_to_keep`.

            In some cases, it may be useful to keep track of a mapping from
            `float -> str` for protected attributes and/or labels. If our use
            case differs from the default, we can modify the mapping stored in
            `metadata`:

            >>> label_map = {1.0: '>50K', 0.0: '<=50K'}
            >>> protected_attribute_maps = [{1.0: 'Male', 0.0: 'Female'}]
            >>> ad = AdultDataset(protected_attribute_names=['sex'],
            ... categorical_features=['workclass', 'education', 'marital-status',
            ... 'occupation', 'relationship', 'native-country', 'race'],
            ... privileged_classes=[['Male']], metadata={'label_map': label_map,
            ... 'protected_attribute_maps': protected_attribute_maps})

            Note that we are now adding `race` as a `categorical_features`.
            Now this information will stay attached to the dataset and can be
            used for more descriptive visualizations.
        """

        train_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                  '..', 'data', 'raw', 'adult', 'adult.data')
        test_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                  '..', 'data', 'raw', 'adult', 'adult.test')
        # as given by adult.names
        column_names = ['age', 'workclass', 'fnlwgt', 'education',
            'education-num', 'marital-status', 'occupation', 'relationship',
            'race', 'sex', 'capital-gain', 'capital-loss', 'hours-per-week',
            'native-country', 'income-per-year']
        try:
            train = pd.read_csv(train_path, header=None, names=column_names,
                skipinitialspace=True, na_values=na_values)
            test = pd.read_csv(test_path, header=0, names=column_names,
                skipinitialspace=True, na_values=na_values)
        except IOError as err:
            print("IOError: {}".format(err))
            print("To use this class, please download the following files:")
            print("\n\thttps://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data")
            print("\thttps://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.test")
            print("\thttps://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.names")
            print("\nand place them, as-is, in the folder:")
            print("\n\t{}\n".format(os.path.abspath(os.path.join(
               os.path.abspath(__file__), '..', '..', 'data', 'raw', 'adult'))))
            import sys
            sys.exit(1)

        df = pd.concat([test, train], ignore_index=True)

        super(AdultDataset, self).__init__(df=df, label_name=label_name,
            favorable_classes=favorable_classes,
            protected_attribute_names=protected_attribute_names,
            privileged_classes=privileged_classes,
            instance_weights_name=instance_weights_name,
            categorical_features=categorical_features,
            features_to_keep=features_to_keep,
            features_to_drop=features_to_drop, na_values=na_values,
            custom_preprocessing=custom_preprocessing, metadata=metadata)



"""
The code for GridSearchReduction wraps the source class
fairlearn.reductions.GridSearch
available in the https://github.com/fairlearn/fairlearn library
licensed under the MIT Licencse, Copyright Microsoft Corporation
"""
try:
    import fairlearn.reductions as red
except ImportError as error:
    from logging import warning
    warning("{}: GridSearchReduction will be unavailable. To install, run:\n"
            "pip install 'aif360[Reductions]'".format(error))
from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.preprocessing import LabelEncoder


class GridSearchReduction(BaseEstimator, ClassifierMixin):
    """Grid search reduction for fair classification or regression.

    Grid search is an in-processing technique that can be used for fair
    classification or fair regression. For classification it reduces fair
    classification to a sequence of cost-sensitive classification problems,
    returning the deterministic classifier with the lowest empirical error
    subject to fair classification constraints [#agarwal18]_ among the
    candidates searched. For regression it uses the same priniciple to return a
    deterministic regressor with the lowest empirical error subject to the
    constraint of bounded group loss [#agarwal19]_.

    References:
        .. [#agarwal18] `A. Agarwal, A. Beygelzimer, M. Dudik, J. Langford, and
           H. Wallach, "A Reductions Approach to Fair Classification,"
           International Conference on Machine Learning, 2018.
           <https://arxiv.org/abs/1803.02453>`_
        .. [#agarwal19] `A. Agarwal, M. Dudik, and Z. Wu, "Fair Regression:
           Quantitative Definitions and Reduction-based Algorithms,"
           International Conference on Machine Learning, 2019.
           <https://arxiv.org/abs/1905.12843>`_
    """
    def __init__(self,
                prot_attr,
                estimator,
                constraints,
                constraint_weight=0.5,
                grid_size=10,
                grid_limit=2.0,
                grid=None,
                drop_prot_attr=True,
                loss="ZeroOne",
                min_val=None,
                max_val=None
                ):
        """
        Args:
            prot_attr: String or array-like column indices or column names
                of protected attributes.
            estimator: An estimator implementing methods ``fit(X, y,
                sample_weight)`` and ``predict(X)``, where ``X`` is the matrix
                of features, ``y`` is the vector of labels, and
                ``sample_weight`` is a vector of weights; labels ``y`` and
                predictions returned by ``predict(X)`` are either 0 or 1 -- e.g.
                scikit-learn classifiers/regressors.
            constraints (str or fairlearn.reductions.Moment): If string, keyword
                denoting the :class:`fairlearn.reductions.Moment` object
                defining the disparity constraints -- e.g., "DemographicParity"
                or "EqualizedOdds". For a full list of possible options see
                `self.model.moments`. Otherwise, provide the desired
                :class:`~fairlearn.reductions.Moment` object defining the
                disparity constraints.
            constraint_weight: When the ``selection_rule`` is
                "tradeoff_optimization" (default, no other option currently)
                this float specifies the relative weight put on the constraint
                violation when selecting the best model. The weight placed on
                the error rate will be ``1-constraint_weight``.
            grid_size (int): The number of Lagrange multipliers to generate in
                the grid.
            grid_limit (float): The largest Lagrange multiplier to generate. The
                grid will contain values distributed between ``-grid_limit`` and
                ``grid_limit`` by default.
            grid (pandas.DataFrame): Instead of supplying a size and limit for
                the grid, users may specify the exact set of Lagrange
                multipliers they desire using this argument in a DataFrame.
            drop_prot_attr (bool): Flag indicating whether to drop protected
                attributes from training data.
            loss (str): String identifying loss function for constraints.
                Options include "ZeroOne", "Square", and "Absolute."
            min_val: Loss function parameter for "Square" and "Absolute,"
                typically the minimum of the range of y values.
            max_val: Loss function parameter for "Square" and "Absolute,"
                typically the maximum of the range of y values.
        """
        self.prot_attr = prot_attr
        self.estimator = estimator
        self.constraints = constraints
        self.constraint_weight = constraint_weight
        self.grid_size = grid_size
        self.grid_limit = grid_limit
        self.grid = grid
        self.drop_prot_attr = drop_prot_attr
        self.loss = loss
        self.min_val = min_val
        self.max_val = max_val

    def fit(self, X, y):
        """Train a less biased classifier or regressor with the given training
        data.

        Args:
            X (pandas.DataFrame): Training samples.
            y (array-like): Training output.

        Returns:
            self
        """
        self.estimator_ = clone(self.estimator)

        moments = {
            "DemographicParity": red.DemographicParity,
            "EqualizedOdds": red.EqualizedOdds,
            "TruePositiveRateParity": red.TruePositiveRateParity,
            "FalsePositiveRateParity": red.FalsePositiveRateParity,
            "ErrorRateParity": red.ErrorRateParity,
            "BoundedGroupLoss": red.BoundedGroupLoss,
        }
        if isinstance(self.constraints, str):
            if self.constraints not in moments:
                raise ValueError(f"Constraint not recognized: {self.constraints}")
            if self.constraints == "BoundedGroupLoss":
                losses = {
                    "ZeroOne": red.ZeroOneLoss,
                    "Square": red.SquareLoss,
                    "Absolute": red.AbsoluteLoss
                }
                if self.loss == "ZeroOne":
                    self.loss_ = losses[self.loss]()
                else:
                    self.loss_ = losses[self.loss](self.min_val, self.max_val)

                self.moment_ = moments[self.constraints](loss=self.loss_)
            else:
                self.moment_ = moments[self.constraints]()
        elif isinstance(self.constraints, red.Moment):
            self.moment_ = self.constraints
        else:
            raise ValueError("constraints must be a string or Moment object.")

        self.model_ = red.GridSearch(estimator=self.estimator_,
                constraints=self.moment_,
                constraint_weight=self.constraint_weight,
                grid_size=self.grid_size, grid_limit=self.grid_limit,
                grid=self.grid)

        A = X[self.prot_attr]

        if self.drop_prot_attr:
            X = X.drop(self.prot_attr, axis=1)

        if isinstance(self.model_.constraints, red.ClassificationMoment):
            le = LabelEncoder()
            y = le.fit_transform(y)
            self.classes_ = le.classes_

        self.model_.fit(X, y, sensitive_features=A)

        return self

    def predict(self, X):
        """Predict output for the given samples.

        Args:
            X (pandas.DataFrame): Test samples.

        Returns:
            numpy.ndarray: Predicted output per sample.
        """
        if self.drop_prot_attr:
            X = X.drop(self.prot_attr, axis=1)

        return self.model_.predict(X)


    def predict_proba(self, X):
        """Probability estimates.

        The returned estimates for all classes are ordered by the label of
        classes for classification.

        Args:
            X (pandas.DataFrame): Test samples.

        Returns:
            numpy.ndarray: returns the probability of the sample for each class
            in the model, where classes are ordered as they are in
            ``self.classes_``.
        """
        if self.drop_prot_attr:
            X = X.drop(self.prot_attr, axis=1)

        if isinstance(self.model_.constraints, red.ClassificationMoment):
            return self.model_.predict_proba(X)

        raise NotImplementedError("Underlying model does not support "
                                  "predict_proba")



try:
    from inFairness import fairalgo
    from skorch import NeuralNet
    from skorch.dataset import unpack_data, Dataset as Dataset_
    from skorch.utils import is_pandas_ndframe
except ImportError as error:
    from logging import warning
    warning("{}: SenSeI and SenSR will be unavailable. To install, run:\n"
            "pip install 'aif360[inFairness]'".format(error))
    Dataset_ = NeuralNet = object
from sklearn.preprocessing import LabelBinarizer
from sklearn.utils.multiclass import type_of_target
from sklearn.exceptions import NotFittedError


class Dataset(Dataset_):
    """A generic dataset class inheriting from :class:`skorch.dataset.Dataset`
    which simply converts DataFrames to arrays instead of treating them like
    dicts.
    """
    def __init__(self, X, y=None, length=None):
        if is_pandas_ndframe(X):
            X = X.values
        if y is not None and is_pandas_ndframe(y):
            y = y.values
        super().__init__(X, y=y, length=length)

class InFairnessNet(NeuralNet):
    """Base class for models which incorporate inFairness algorithms."""
    def __init__(self, *args, criterion, train_split=None, regression='auto',
                 dataset=Dataset, **kwargs):
        """
        Args:
            criterion (torch.nn.Module, keyword-only): Loss function.
            train_split (callable, optional): See :class:`skorch.net.NeuralNet`.
                Note: validation loss *does not* include any fairness loss, only
                the provided criterion, and should not be used for early
                stopping, etc. Default is None (no split).
            regression (bool or 'auto'): Task is regression. If 'auto', this is
                inferred using :func:`sklearn.utils.multiclass.type_of_target`
                on y in fit(). If a Dataset is provided to fit, this defaults to
                False. If y contains 'soft' targets (i.e. probabilities per
                class), this should be manually set to False.
        """
        super().__init__(*args, criterion=criterion, train_split=train_split,
                         dataset=dataset, **kwargs)
        self.regression = regression

    @property
    def _estimator_type(self):
        if hasattr(self, "regression_"):
            return 'regressor' if self.regression_ else 'classifier'
        elif self.regression != 'auto':
            return 'regressor' if self.regression else 'classifier'
        else:
            raise NotFittedError("regression is set to 'auto'. Call 'fit' with "
                    "appropriate arguments or set regression manually.")

    def get_loss(self, y_pred, y_true, X=None, training=False):
        """Return the loss for this batch.

        Parameters
        ----------
        y_pred : torch tensor
          Predicted target values

        y_true : torch tensor
          True target values.

        X : input data, compatible with skorch.dataset.Dataset
          By default, you should be able to pass:

            * numpy arrays
            * torch tensors
            * pandas DataFrame or Series
            * scipy sparse CSR matrices
            * a dictionary of the former three
            * a list/tuple of the former three
            * a Dataset

          If this doesn't work with your data, you have to pass a
          ``Dataset`` that can deal with the data.

        training : bool (default=False)
          Whether train mode should be used or not.

        """
        if training:
            return y_pred.loss
        else:
            return super().get_loss(y_pred.y_pred, y_true)

    def train_step_single(self, batch, **fit_params):
        """Compute y_pred, loss value, and update net's gradients.

        The module is set to be in train mode (e.g. dropout is
        applied).

        Parameters
        ----------
        batch
          A single batch returned by the data loader.

        **fit_params : dict
          Additional parameters passed to the ``forward`` method of
          the module and to the ``self.train_split`` call.

        Returns
        -------
        step : dict
          A dictionary ``{'loss': loss, 'y_pred': y_pred}``, where the
          float ``loss`` is the result of the loss function and
          ``y_pred`` the prediction generated by the PyTorch module.

        """
        self._set_training(True)
        Xi, yi = unpack_data(batch)
        response = self.infer(Xi, Y=yi, **fit_params)
        loss = self.get_loss(response, yi, X=Xi, training=True)
        loss.backward()
        return {
            'loss': loss,
            'y_pred': response.y_pred,
        }

    def validation_step(self, batch, **fit_params):
        """Perform a forward step using batched data and return the
        resulting loss.

        The module is set to be in evaluation mode (e.g. dropout is
        not applied).

        Parameters
        ----------
        batch
          A single batch returned by the data loader.

        **fit_params : dict
          Additional parameters passed to the ``forward`` method of
          the module and to the ``self.train_split`` call.

        """
        step = super().validation_step(batch, **fit_params)
        return {'loss': step['loss'], 'y_pred': step['y_pred'].y_pred}

    def evaluation_step(self, batch, training=False):
        """Perform a forward step to produce the output used for
        prediction and scoring.

        Therefore, the module is set to evaluation mode by default
        beforehand which can be overridden to re-enable features
        like dropout by setting ``training=True``.

        Parameters
        ----------
        batch
          A single batch returned by the data loader.

        training : bool (default=False)
          Whether to set the module to train mode or not.

        Returns
        -------
        y_infer
          The prediction generated by the module.

        """
        return super().evaluation_step(batch, training=training).y_pred

    def fit(self, X, y, **fit_params):
        """Initialize and fit the model.

        If the module was already initialized, by calling fit, the module will
        be re-initialized (unless ``warm_start`` is True).

        Args:
            X (array-like): Training samples.
            y (array-like): Training labels. If X is a Dataset that contains
                the target, y may be set to ``None``. Note: if ``regression`` is
                'auto' in this case, classification will be assumed. If y is
                already binary/ordinal encoded (i.e., the unique labels consist
                of the integers from [0, C)) it is passed as-is, however, if y
                contains nominal values, it is encoded with
                :class:`sklearn.preprocessing.LabelBinarizer` and cast to
                'float32' for compatibility with torch loss functions.
                Regression targets are also left unmodified.
            **fit_params: Additional parameters passed to the `forward`` method
                of the module and to the ``self.train_split`` call.

        Returns:
            self
        """
        self.regression_ = self.regression
        if y is not None and self.regression_ is not True:
            ttype = type_of_target(y)
            if ttype in ("binary", "multiclass", "multilabel-indicator"):
                lb = LabelBinarizer().fit(y)
                self.classes_ = lb.classes_
                if self.classes_.tolist() != list(range(len(self.classes_))):
                    y = lb.transform(y).astype('float32')
            elif "continuous" in ttype and self.regression_ == 'auto':
                self.regression_ = True
            else:
                raise ValueError(f'Detected {ttype} type y with regression='
                    f'{self.regression}. This combination is not supported.')
        if self.regression_ == 'auto':
            self.regression_ = False
        return super().fit(X, y, **fit_params)

    def predict(self, X):
        """Return class labels for samples in X if task is classification or
        predicted values if task is regression.

        Args:
            X (array-like): Test samples.

        Returns:
            numpy.ndarray: Test predictions.

        See also:
            :meth:`skorch.net.NeuralNet.predict`
        """
        if self.regression_:
            return super().predict(X)
        elif hasattr(self, "classes_"):
            return self.classes_[self.predict_proba(X).argmax(axis=1)]
        else:
            return self.predict_proba(X).argmax(axis=1)

class SenSeI(InFairnessNet):
    """Sensitive Set Invariance (SenSeI).

    SenSeI is an in-processing method for individual fairness [#yurochkin20]_.
    In this method, individual fairness is formulated as invariance on certain
    sensitive sets. SenSeI minimizes a transport-based regularizer that enforces
    this version of individual fairness.

    References:
        .. [#yurochkin20] `M. Yurochkin and Y. Sun, "SenSeI: Sensitive Set
           Invariance for Enforcing Individual Fairness." International
           Conference on Learning Representations, 2021.
           <https://arxiv.org/abs/2006.14168>`_

    See also:
        :class:`inFairness.fairalgo.SenSeI`

    Attributes:
        regression_ (bool): Whether or not this task is treated as regression.
        classes_ (array, shape (n_classes,)): A list of class labels known to
            the transformer. Only present if ``self.regression_`` is False and
            y is provided to ``fit``.
        module_ (inFairness.fairalgo.SenSeI): The fair PyTorch module.
    """
    def __init__(self, module, *, criterion, distance_x, distance_y, rho, eps,
                 auditor_nsteps, auditor_lr, regression='auto', **kwargs):
        """
        Args:
            module (torch.nn.Module): Network architecture.
            criterion (torch.nn.Module): Loss function.
            distance_x (inFairness.distances.Distance): Distance metric in the
                input space.
            distance_y (inFairness.distances.Distance): Distance metric in the
                output space.
            rho (float): :math:`\\rho` parameter in the SenSeI algorithm.
            eps (float): :math:`\epsilon` parameter in the SenSeI algorithm.
            auditor_nsteps (int): Number of update steps for the auditor to find
                worst-case examples
            auditor_lr (float): Learning rate for the auditor.
            regression (bool or 'auto'): Task is regression. If 'auto', this is
                inferred using :func:`sklearn.utils.multiclass.type_of_target`
                on y in fit(). If a Dataset is provided to fit, this defaults to
                False. If y contains 'soft' targets (i.e. probabilities per
                class), this should be manually set to False.
            train_split (callable, optional): See :class:`skorch.net.NeuralNet`.
                Note: validation loss *does not* include any fairness loss, only
                the provided criterion, and should not be used for early
                stopping, etc. Default is None (no split).
            **kwargs: See :class:`skorch.net.NeuralNet`.
        """
        self.distance_x = distance_x
        self.distance_y = distance_y
        self.rho = rho
        self.eps = eps
        self.auditor_nsteps = auditor_nsteps
        self.auditor_lr = auditor_lr

        super().__init__(module=module, criterion=criterion,
                         regression=regression, **kwargs)

    def initialize_module(self):
        """Initializes the module.

        If the module is already initialized and no parameter was changed, it
        will be left as is.
        """
        self.initialize_criterion()
        kwargs = self.get_params_for('module')
        network = self.initialized_instance(self.module, kwargs)

        sensei_kwargs = {
            'network': network,
            'loss_fn': self.criterion_,
            'distance_x': self.distance_x,
            'distance_y': self.distance_y,
            'rho': self.rho,
            'eps': self.eps,
            'auditor_nsteps': self.auditor_nsteps,
            'auditor_lr': self.auditor_lr,
        }
        self.module_ = self.initialized_instance(fairalgo.SenSeI, sensei_kwargs)
        return self

class SenSR(InFairnessNet):
    """Sensitive Subspace Robustness (SenSR).

    SenSR is an in-processing method for individual fairness which enforces
    performance invariance under certain sensitive perturbations to the input
    [#yurochkin19]_.

    References:
        .. [#yurochkin19] `M. Yurochkin, A. Bower, and Y. Sun, "Training
           individually fair ML models with sensitive subspace robustness."
           International Conference on Learning Representations, 2020.
           <https://arxiv.org/abs/1907.00020>`_

    See also:
        :class:`inFairness.fairalgo.SenSR`

    Attributes:
        regression_ (bool): Whether or not this task is treated as regression.
        classes_ (array, shape (n_classes,)): A list of class labels known to
            the transformer. Only present if ``self.regression_`` is False and
            y is provided to ``fit``.
        module_ (inFairness.fairalgo.SenSR): The fair PyTorch module.
    """
    def __init__(self, module, *, criterion, distance_x, eps, lr_lamb, lr_param,
                 auditor_nsteps, auditor_lr, regression='auto', **kwargs):
        """
        Args:
            module (torch.nn.Module): Network architecture.
            criterion (torch.nn.Module): Loss function.
            distance_x (inFairness.distances.Distance): Distance metric in the
                input space.
            eps (float): :math:`\epsilon` parameter in the SenSR algorithm.
            lr_lamb (float): :math:`\lambda` parameter in the SenSR algorithm.
            lr_param (float): :math:`\\alpha` parameter in the SenSR algorithm.
            auditor_nsteps (int): Number of update steps for the auditor to find
                worst-case examples
            auditor_lr (float): Learning rate for the auditor.
            regression (bool or 'auto'): Task is regression. If 'auto', this is
                inferred using :func:`sklearn.utils.multiclass.type_of_target`
                on y in fit(). If a Dataset is provided to fit, this defaults to
                False. If y contains 'soft' targets (i.e. probabilities per
                class), this should be manually set to False.
            train_split (callable, optional): See :class:`skorch.net.NeuralNet`.
                Note: validation loss *does not* include any fairness loss, only
                the provided criterion, and should not be used for early
                stopping, etc. Default is None (no split).
            **kwargs: See :class:`skorch.net.NeuralNet`.
        """
        self.distance_x = distance_x
        self.eps = eps
        self.lr_lamb = lr_lamb
        self.lr_param = lr_param
        self.auditor_nsteps = auditor_nsteps
        self.auditor_lr = auditor_lr

        super().__init__(module=module, criterion=criterion,
                         regression=regression, **kwargs)

    def initialize_module(self):
        """Initializes the module.

        If the module is already initialized and no parameter was changed, it
        will be left as is.
        """
        self.initialize_criterion()
        kwargs = self.get_params_for('module')
        network = self.initialized_instance(self.module, kwargs)

        sensr_kwargs = {
            'network': network,
            'loss_fn': self.criterion_,
            'distance_x': self.distance_x,
            'eps': self.eps,
            'lr_lamb': self.lr_lamb,
            'lr_param': self.lr_param,
            'auditor_nsteps': self.auditor_nsteps,
            'auditor_lr': self.auditor_lr,
        }
        self.module_ = self.initialized_instance(fairalgo.SenSR, sensr_kwargs)
        return self



from functools import partial

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
import pandas as pd
from pandas.api.types import is_numeric_dtype
from pandas.testing import assert_frame_equal
import pytest
from sklearn.compose import make_column_transformer
from sklearn.preprocessing import OneHotEncoder, minmax_scale

from aif360.datasets import (
    AdultDataset, GermanDataset, CompasDataset, LawSchoolGPADataset,
    MEPSDataset19, MEPSDataset20, MEPSDataset21)
from aif360.sklearn.datasets import (
    standardize_dataset, NumericConversionWarning, fetch_adult, fetch_bank,
    fetch_german, fetch_compas, fetch_lawschool_gpa, fetch_meps)


df = pd.DataFrame([[1, 2, 3, 'a'], [5, 6, 7, 'b'], [np.NaN, 10, 11, 'c']],
                  columns=['X1', 'X2', 'y', 'Z'])
basic = partial(standardize_dataset, df=df, prot_attr='Z', target='y',
                dropna=False)

def test_standardize_dataset_basic():
    """Tests standardize_dataset on a toy example."""
    dataset = basic()
    X, y = dataset
    X, y = dataset.X, dataset.y
    with pytest.raises(ValueError):
        X, y, sample_weight = dataset
    with pytest.raises(AttributeError):
        dataset.sample_weight
    assert isinstance(X, pd.DataFrame)
    assert isinstance(y, pd.Series)
    assert X.index.equals(y.index)
    assert X.shape == (3, 3)

def test_multilabel_basic():
    """Tests returning a multilabel target on a toy example."""
    multilabel = basic(target=['X2', 'y'])
    assert isinstance(multilabel.y, pd.DataFrame)
    assert isinstance(multilabel.X, pd.DataFrame)
    assert multilabel.y.shape == (3, 2)
    assert multilabel.X.shape == (3, 2)

def test_sample_weight_basic():
    """Tests returning sample_weight on a toy example."""
    with_weights = basic(sample_weight='X2')
    assert len(with_weights) == 3
    assert with_weights.X.shape == (3, 2)

def test_array_args_basic():
    """Tests passing explicit arrays instead of column labels for prot_attr,
    target, and sample_weight.
    """
    # single array
    pa_array = basic(prot_attr=pd.Index([1, 0, 1], name='ZZ'))
    assert pa_array.X.columns.equals(pd.Index(['X1', 'X2', 'Z']))
    assert pa_array.X.index.names == ['ZZ']
    # mixed array and label
    tar_array_mixed = basic(target=[np.array([4, 8, 12]), 'y'])
    assert tar_array_mixed.y.shape == (3, 2)
    assert tar_array_mixed.X.shape == (3, 3)
    assert tar_array_mixed.y.index.equals(tar_array_mixed.X.index)
    # sample weight
    sw_array = basic(sample_weight=[0.5, 0.4, 2.1])
    assert sw_array.sample_weight.index.equals(sw_array.X.index)

def test_usecols_dropcols_basic():
    """Tests various combinations of usecols and dropcols on a toy example."""
    assert basic(usecols=['X1']).X.columns.tolist() == ['X1']
    assert basic(usecols=['X1', 'Z']).X.columns.tolist() == ['X1', 'Z']

    assert basic(dropcols=['X1']).X.columns.tolist() == ['X2', 'Z']
    assert basic(dropcols=['X1', 'Z']).X.columns.tolist() == ['X2']

    assert basic(usecols=['X1'], dropcols=['X2']).X.columns.tolist() == ['X1']
    assert isinstance(basic(usecols=['X2'], dropcols=['X1', 'X2'])[0],
                      pd.DataFrame)

def test_dropna_basic():
    """Tests dropna on a toy example."""
    basic_dropna = partial(standardize_dataset, df=df, prot_attr='Z',
                           target='y', dropna=True)
    assert basic_dropna().X.shape == (2, 3)
    assert basic(dropcols=['X1']).X.shape == (3, 2)

@pytest.mark.filterwarnings('ignore', category=NumericConversionWarning)
def test_numeric_only_basic():
    """Tests numeric_only on a toy example."""
    num_only = basic(numeric_only=True)
    assert num_only.X.shape == (3, 2)
    assert 'Z' in num_only.X.index.names
    num_only_X2 = basic(prot_attr='X2', numeric_only=True)
    num_only_X2_dropZ = basic(prot_attr='X2', dropcols=['Z'], numeric_only=True)
    assert num_only_X2.X.equals(num_only_X2_dropZ.X)

@pytest.mark.filterwarnings('error', category=NumericConversionWarning)
def test_numeric_only_warnings():
    with pytest.raises(UserWarning):
        basic(numeric_only=True)  # prot_attr has non-numeric
    with pytest.raises(UserWarning):
        basic(numeric_only=True, prot_attr='y', target='Z')  # y has non-numeric

def test_multiindex_cols():
    """Tests DataFrame with MultiIndex columns."""
    cols = pd.MultiIndex.from_arrays([['X', 'X', 'y', 'Z'], [1, 2, '', '']])
    df = pd.DataFrame([[1, 2, 3, 'a'], [5, 6, 7, 'b'], [None, 10, 11, 'c']],
                    columns=cols)
    multiindex = standardize_dataset(df, prot_attr='Z', target='y')
    assert multiindex.X.index.names == ['Z']
    assert multiindex.y.name == 'y'
    assert multiindex.X.columns.equals(cols.drop('y', level=0))

@pytest.mark.filterwarnings('ignore', category=NumericConversionWarning)
def test_fetch_adult():
    """Tests Adult Income dataset shapes with various options."""
    adult = fetch_adult()
    assert len(adult) == 3
    assert adult.X.shape == (45222, 13)
    assert len(adult.X.index.get_level_values('race').categories) == 2
    assert len(adult.X.race.cat.categories) > 2
    assert fetch_adult(dropna=False).X.shape == (48842, 13)
    # race is kept since it's binary
    assert fetch_adult(numeric_only=True).X.shape == (48842, 7)
    num_only_bin_race = fetch_adult(numeric_only=True, binary_race=False)
    # race gets dropped since it's categorical
    assert num_only_bin_race.X.shape == (48842, 6)
    # still in index though
    assert 'race' in num_only_bin_race.X.index.names

def test_adult_matches_old():
    """Tests Adult Income dataset matches original version."""
    X, y, _ = fetch_adult()
    X.race = X.race.cat.set_categories(['Non-white', 'White']).fillna('Non-white')

    adult = AdultDataset()
    adult = adult.convert_to_dataframe(de_dummy_code=True)[0].drop(columns=adult.label_names)

    assert_frame_equal(X.reset_index(drop=True), adult.reset_index(drop=True),
                       check_dtype=False, check_categorical=False, check_like=True)

def test_fetch_german():
    """Tests German Credit dataset shapes with various options."""
    german = fetch_german()
    assert len(german) == 2
    assert german.X.shape == (1000, 21)
    assert fetch_german(numeric_only=True).X.shape == (1000, 9)

def test_german_matches_old():
    """Tests German Credit datasets matches original version."""
    column_map = {
        'checking_status': 'status',
        'duration': 'month',
        'savings_status': 'savings',
        'installment_commitment': 'investment_as_income_percentage',
        'other_parties': 'other_debtors',
        'property_magnitude': 'property',
        'other_payment_plans': 'installment_plans',
        'existing_credits': 'number_of_credits',
        'job': 'skill_level',
        'num_dependents': 'people_liable_for',
        'own_telephone': 'telephone',
    }
    X, y = fetch_german()
    # marital status was not included before and age was binary
    X = X.drop(columns=['marital_status', 'age']).reset_index('age')
    # columns are named differently in the old version
    X = X.rename(columns=column_map)

    old = GermanDataset()
    old = old.convert_to_dataframe(de_dummy_code=True)[0].drop(columns=old.label_names)

    # categories in the old version were not renamed so just map both to ints
    X = X.apply(lambda c: c.factorize()[0] if not is_numeric_dtype(c) else c)
    old = old.apply(lambda c: c.factorize()[0] if not is_numeric_dtype(c) else c)

    assert_frame_equal(X.reset_index(drop=True), old.reset_index(drop=True),
                       check_like=True, check_dtype=False)

def test_fetch_bank():
    """Tests Bank Marketing dataset shapes with various options."""
    bank = fetch_bank()
    assert len(bank) == 2
    assert bank.X.shape == (45211, 15)
    assert fetch_bank(dropcols=None).X.shape == (45211, 16)
    assert fetch_bank(numeric_only=True).X.shape == (45211, 7)

# TODO: bank doesn't match old

@pytest.mark.filterwarnings('ignore', category=NumericConversionWarning)
def test_fetch_compas():
    """Tests COMPAS Recidivism dataset shapes with various options."""
    compas = fetch_compas()
    assert len(compas) == 2
    assert compas.X.shape == (6167, 10)
    assert fetch_compas(binary_race=True).X.shape == (5273, 10)
    assert fetch_compas(numeric_only=True).X.shape == (6172, 8)
    assert fetch_compas(numeric_only=True, binary_race=True).X.shape == (5278, 9)

def test_compas_matches_old():
    """Tests COMPAS Recidivism dataset matches original version."""
    X, y = fetch_compas()
    X.race = X.race.cat.set_categories(['Not Caucasian', 'Caucasian']).fillna('Not Caucasian')

    compas = CompasDataset()
    compas = compas.convert_to_dataframe(de_dummy_code=True)[0].drop(columns=compas.label_names)

    assert_frame_equal(X.reset_index(drop=True), compas.reset_index(drop=True),
                       check_dtype=False, check_categorical=False, check_like=True)

def test_fetch_lawschool_gpa():
    """Tests Law School GPA dataset shapes with various options."""
    gpa = fetch_lawschool_gpa()
    assert len(gpa) == 2
    assert gpa.X.shape == (22342, 4)
    assert gpa.y.nunique() > 2  # regression
    assert fetch_lawschool_gpa(numeric_only=True, dropna=False).X.shape == (27478, 4)

def test_lawschool_matches_old():
    """Tests Law School GPA dataset matches original version."""
    X, y = fetch_lawschool_gpa(numeric_only=True)

    law = LawSchoolGPADataset()
    law = law.convert_to_dataframe()[0].drop(columns=law.label_names)

    assert_array_equal(minmax_scale(X), law)

@pytest.mark.parametrize("panel", [19, 20, 21])
def test_cache_meps(panel):
    """Tests if cached MEPS matches raw."""
    meps_raw = fetch_meps(panel, cache=False, accept_terms=True)[0]
    fetch_meps(panel, cache=True, accept_terms=True)
    meps_cached = fetch_meps(panel, cache=True)[0]
    assert_frame_equal(meps_raw, meps_cached)
    assert_array_equal(meps_raw.to_numpy(), meps_cached.to_numpy())

@pytest.mark.parametrize(
    "panel, cls",
    [(19, MEPSDataset19), (20, MEPSDataset20), (21, MEPSDataset21)])
def test_meps_matches_old(panel, cls):
    """Tests MEPS datasets match original versions."""
    usecols = ['REGION', 'AGE', 'SEX', 'RACE', 'MARRY', 'FTSTU',
               'ACTDTY', 'HONRDC', 'RTHLTH', 'MNHLTH', 'HIBPDX',
               'CHDDX', 'ANGIDX', 'MIDX', 'OHRTDX', 'STRKDX', 'EMPHDX',
               'CHBRON', 'CHOLDX', 'CANCERDX', 'DIABDX', 'JTPAIN',
               'ARTHDX', 'ARTHTYPE', 'ASTHDX', 'ADHDADDX', 'PREGNT',
               'WLKLIM', 'ACTLIM', 'SOCLIM', 'COGLIM', 'DFHEAR42',
               'DFSEE42', 'ADSMOK42', 'PCS42', 'MCS42', 'K6SUM42',
               'PHQ242', 'EMPST', 'POVCAT', 'INSCOV']
    educols = ['EDUCYR', 'HIDEG']
    meps = fetch_meps(panel, accept_terms=True, usecols=usecols + educols)
    assert len(meps) == 3
    meps.X.RACE = meps.X.RACE.factorize(sort=True)[0]
    MEPS = cls()
    assert_allclose(pd.get_dummies(meps.X.drop(columns=educols)).astype(float),
                    MEPS.features, atol=1e-16)
    assert_array_equal(meps.y.factorize(sort=True)[0], MEPS.labels.ravel())

@pytest.mark.parametrize("panel", [19, 20, 21])
def test_fetch_meps(panel):
    """Tests MEPS datasets shapes with various options."""
    meps = fetch_meps(panel, accept_terms=True, dropna=False)
    meps_dropna = fetch_meps(panel, dropna=True)
    assert meps_dropna.X.shape[0] < meps.X.shape[0]
    meps_numeric = fetch_meps(panel, accept_terms=True, numeric_only=True)
    assert meps_numeric.X.shape[1] == 5

def test_onehot_transformer():
    """Tests that categorical features can be correctly one-hot encoded."""
    X, y = fetch_german()
    ohe = make_column_transformer(
        (OneHotEncoder(), X.dtypes == 'category'),
        remainder='passthrough', verbose_feature_names_out=False)
    dum = pd.get_dummies(X)
    assert ohe.fit_transform(X).shape[1] == dum.shape[1] == 63
    assert dum.columns.symmetric_difference(ohe.get_feature_names_out()).empty



import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_list_like
from sklearn.utils import check_consistent_length
from sklearn.utils.validation import column_or_1d


def check_inputs(X, y, sample_weight=None, ensure_2d=True):
    """Input validation for debiasing algorithms.

    Checks all inputs for consistent length, validates shapes (optional for X),
    and returns an array of all ones if sample_weight is ``None``.

    Args:
        X (array-like): Input data.
        y (array-like, shape = (n_samples,)): Target values.
        sample_weight (array-like, optional): Sample weights.
        ensure_2d (bool, optional): Whether to raise a ValueError if X is not
            2D.

    Returns:
        tuple:

            * **X** (`array-like`) -- Validated X. Unchanged.

            * **y** (`array-like`) -- Validated y. Possibly converted to 1D if
              not a :class:`pandas.Series`.
            * **sample_weight** (`array-like`) -- Validated sample_weight. If no
              sample_weight is provided, returns a consistent-length array of
              ones.
    """
    if ensure_2d and X.ndim != 2:
        raise ValueError("Expected X to be 2D, got ndim == {} instead.".format(
                X.ndim))
    if not isinstance(y, pd.Series):  # don't cast Series -> ndarray
        y = column_or_1d(y)
    if sample_weight is not None:
        sample_weight = column_or_1d(sample_weight)
    else:
        sample_weight = np.ones(X.shape[0])
    check_consistent_length(X, y, sample_weight)
    return X, y, sample_weight

def check_groups(arr, prot_attr, ensure_binary=False):
    """Get groups from the index of arr.

    If there are multiple protected attributes provided, the index is flattened
    to be a 1-D Index of tuples. If ensure_binary is ``True``, raises a
    ValueError if there are not exactly two unique groups. Also checks that all
    provided protected attributes are in the index.

    Args:
        arr (array-like): Either a Pandas object containing protected attribute
            information in the index or array-like with explicit protected
            attribute array(s) for `prot_attr`.
        prot_attr (label or array-like or list of labels/arrays): Protected
            attribute(s). If contains labels, arr must include these in its
            index. If ``None``, all protected attributes in ``arr.index`` are
            used. Can also be 1D array-like of the same length as arr or a
            list of a combination of such arrays and labels in which case, arr
            may not necessarily be a Pandas type.
        ensure_binary (bool): Raise an error if the resultant groups are not
            binary.

    Returns:
        tuple:

            * **groups** (:class:`pandas.Index`) -- Label (or tuple of labels)
              of protected attribute for each sample in arr.
            * **prot_attr** (`FrozenList`) -- Modified input. If input is a
              single label, returns single-item list. If input is ``None``
              returns list of all protected attributes.
    """
    arr_is_pandas = isinstance(arr, (pd.DataFrame, pd.Series))
    if prot_attr is None:  # use all protected attributes provided in arr
        if not arr_is_pandas:
            raise TypeError("Expected `Series` or `DataFrame` for arr, got "
                           f"{type(arr).__name__} instead. Otherwise, pass "
                            "explicit prot_attr array(s).")
        groups = arr.index
    elif arr_is_pandas:
        df = arr.index.to_frame()
        groups = df.set_index(prot_attr).index  # let pandas handle errors
    else:  # arr isn't pandas. might be okay if prot_attr is array-like
        df = pd.DataFrame(index=[None]*len(arr))  # dummy to check lengths match
        try:
            groups = df.set_index(prot_attr).index
        except KeyError as e:
            raise TypeError("arr does not include protected attributes in the "
                            "index. Check if this got dropped or prot_attr is "
                            "formatted incorrectly.") from e
    prot_attr = groups.names
    groups = groups.to_flat_index()

    n_unique = groups.nunique()
    if ensure_binary and n_unique != 2:
        raise ValueError("Expected 2 protected attribute groups, got "
                        f"{groups.unique() if n_unique > 5 else n_unique}")

    return groups, prot_attr



import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted

from aif360.sklearn.metrics import difference, base_rate
from aif360.sklearn.metrics import generalized_fnr, generalized_fpr
from aif360.sklearn.utils import check_inputs, check_groups


class CalibratedEqualizedOdds(BaseEstimator, ClassifierMixin):
    """Calibrated equalized odds post-processor.

    Calibrated equalized odds is a post-processing technique that optimizes over
    calibrated classifier score outputs to find probabilities with which to
    change output labels with an equalized odds objective [#pleiss17]_.

    Note:
        A :class:`~sklearn.pipeline.Pipeline` expects a single estimation step
        but this class requires an estimator's predictions as input. See
        :class:`PostProcessingMeta` for a workaround.

    See also:
        :class:`PostProcessingMeta`

    References:
        .. [#pleiss17] `G. Pleiss, M. Raghavan, F. Wu, J. Kleinberg, and
           K. Q. Weinberger, "On Fairness and Calibration," Conference on Neural
           Information Processing Systems, 2017.
           <http://papers.nips.cc/paper/7151-on-fairness-and-calibration.pdf>`_

    Adapted from:
    https://github.com/gpleiss/equalized_odds_and_calibration/blob/master/calib_eq_odds.py

    Attributes:
        prot_attr_ (str or list(str)): Protected attribute(s) used for post-
            processing.
        groups_ (array, shape (2,)): A list of group labels known to the
            classifier. Note: this algorithm require a binary division of the
            data.
        classes_ (array, shape (num_classes,)): A list of class labels known to
            the classifier. Note: this algorithm treats all non-positive
            outcomes as negative (binary classification only).
        pos_label_ (scalar): The label of the positive class.
        mix_rates_ (array, shape (2,)): The interpolation parameters -- the
            probability of randomly returning the group's base rate. The group
            for which the cost function is higher is set to 0.
    """
    def __init__(self, prot_attr=None, cost_constraint='weighted',
                 random_state=None):
        """
        Args:
            prot_attr (single label or list-like, optional): Protected
                attribute(s) to use in the post-processing. If more than one
                attribute, all combinations of values (intersections) are
                considered. Default is ``None`` meaning all protected attributes
                from the dataset are used. Note: This algorithm requires there
                be exactly 2 groups (privileged and unprivileged).
            cost_constraint ('fpr', 'fnr', or 'weighted'): Which equal-cost
                constraint to satisfy: generalized false positive rate ('fpr'),
                generalized false negative rate ('fnr'), or a weighted
                combination of both ('weighted').
            random_state (int or numpy.RandomState, optional): Seed of pseudo-
                random number generator for sampling from the mix rates.
        """
        self.prot_attr = prot_attr
        self.cost_constraint = cost_constraint
        self.random_state = random_state

    def _more_tags(self):
        return {'requires_proba': True}

    def _weighted_cost(self, y_true, probas_pred, pos_label=1,
                       sample_weight=None):
        """Evaluates the cost function specified by ``self.cost_constraint``."""
        if self.cost_constraint == 'fpr':
            return generalized_fpr(y_true, probas_pred, pos_label=pos_label,
                                   sample_weight=sample_weight)
        elif self.cost_constraint == 'fnr':
            return generalized_fnr(y_true, probas_pred, pos_label=pos_label,
                                   sample_weight=sample_weight)
        elif self.cost_constraint == 'weighted':
            fpr = generalized_fpr(y_true, probas_pred, pos_label=pos_label,
                                  sample_weight=sample_weight)
            fnr = generalized_fnr(y_true, probas_pred, pos_label=pos_label,
                                  sample_weight=sample_weight)
            br = base_rate(y_true, probas_pred, pos_label=pos_label,
                           sample_weight=sample_weight)
            return fpr * (1 - br) + fnr * br
        else:
            raise ValueError("`cost_constraint` must be one of: 'fpr', 'fnr', "
                             "or 'weighted'")

    def fit(self, X, y, labels=None, pos_label=1, sample_weight=None):
        """Compute the mixing rates required to satisfy the cost constraint.

        Args:
            X (array-like): Probability estimates of the targets as returned by
                a ``predict_proba()`` call or equivalent.
            y (pandas.Series): Ground-truth (correct) target values.
            labels (list, optional): The ordered set of labels values. Must
                match the order of columns in X if provided. By default,
                all labels in y are used in sorted order.
            pos_label (scalar, optional): The label of the positive class.
            sample_weight (array-like, optional): Sample weights.

        Returns:
            self
        """
        X, y, sample_weight = check_inputs(X, y, sample_weight)
        groups, self.prot_attr_ = check_groups(y, self.prot_attr,
                                               ensure_binary=True)
        self.classes_ = np.array(labels) if labels is not None else np.unique(y)
        self.groups_ = np.unique(groups)
        self.pos_label_ = pos_label

        if len(self.classes_) != 2:
            raise ValueError('Only binary classification is supported.')
        if len(self.classes_) != X.shape[1]:
            raise ValueError('Only binary classification is supported. X should'
                    ' contain one column per class. Got: {} columns.'.format(
                            X.shape[1]))

        if pos_label not in self.classes_:
            raise ValueError('pos_label={} is not in the set of labels. The '
                    'valid values are:\n{}'.format(pos_label, self.classes_))

        pos_idx = np.nonzero(self.classes_ == self.pos_label_)[0][0]
        try:
            X = X.iloc[:, pos_idx]
        except AttributeError:
            X = X[:, pos_idx]

        # local function to evaluate corresponding metric
        def _eval(func, grp_idx, trivial=False):
            idx = (groups == self.groups_[grp_idx])
            pred = np.full_like(X, self.base_rates_[grp_idx]) if trivial else X
            return func(y[idx], pred[idx], pos_label=pos_label,
                        sample_weight=sample_weight[idx])

        self.base_rates_ = [_eval(base_rate, i) for i in range(2)]

        costs = np.array([[_eval(self._weighted_cost, i, t) for i in range(2)]
                          for t in (False, True)])
        self.mix_rates_ = [
                (costs[0, 1] - costs[0, 0]) / (costs[1, 0] - costs[0, 0]),
                (costs[0, 0] - costs[0, 1]) / (costs[1, 1] - costs[0, 1])]
        self.mix_rates_[np.argmax(costs[0])] = 0

        return self

    def predict_proba(self, X):
        """The returned estimates for all classes are ordered by the label of
        classes.

        Args:
            X (pandas.DataFrame): Probability estimates of the targets as
                returned by a ``predict_proba()`` call or equivalent. Note: must
                include protected attributes in the index.

        Returns:
            numpy.ndarray: Returns the probability of the sample for each class
            in the model, where classes are ordered as they are in
            ``self.classes_``.
        """
        check_is_fitted(self, 'mix_rates_')
        rng = check_random_state(self.random_state)

        groups, _ = check_groups(X, self.prot_attr_)
        if not set(np.unique(groups)) <= set(self.groups_):
            raise ValueError('The protected groups from X:\n{}\ndo not '
                             'match those from the training set:\n{}'.format(
                                     np.unique(groups), self.groups_))

        pos_idx = np.nonzero(self.classes_ == self.pos_label_)[0][0]
        X = X.iloc[:, pos_idx]

        yt = np.empty_like(X)
        for grp_idx in range(2):
            i = (groups == self.groups_[grp_idx])
            to_replace = (rng.rand(sum(i)) < self.mix_rates_[grp_idx])
            new_preds = X[i].copy()
            new_preds[to_replace] = self.base_rates_[grp_idx]
            yt[i] = new_preds

        return np.c_[1 - yt, yt] if pos_idx == 1 else np.c_[yt, 1 - yt]

    def predict(self, X):
        """Predict class labels for the given scores.

        Args:
            X (pandas.DataFrame): Probability estimates of the targets as
                returned by a ``predict_proba()`` call or equivalent. Note: must
                include protected attributes in the index.

        Returns:
            numpy.ndarray: Predicted class label per sample.
        """
        scores = self.predict_proba(X)
        return self.classes_[scores.argmax(axis=1)]

    def score(self, X, y, sample_weight=None):
        """Score the predictions according to the cost constraint specified.

        Args:
            X (pandas.DataFrame): Probability estimates of the targets as
                returned by a ``predict_proba()`` call or equivalent. Note: must
                include protected attributes in the index.
            y (array-like): Ground-truth (correct) target values.
            sample_weight (array-like, optional): Sample weights.

        Returns:
            float: Absolute value of the difference in cost function for the two
            groups (e.g. :func:`~aif360.sklearn.metrics.generalized_fpr` if
            ``self.cost_constraint`` is 'fpr')
        """
        check_is_fitted(self, ['classes_', 'pos_label_'])
        pos_idx = np.nonzero(self.classes_ == self.pos_label_)[0][0]
        probas_pred = self.predict_proba(X)[:, pos_idx]

        return abs(difference(self._weighted_cost, y, probas_pred,
                prot_attr=self.prot_attr_, priv_group=self.groups_[1],
                pos_label=self.pos_label_, sample_weight=sample_weight))



"""
Pre-processing algorithms modify a dataset to be more fair (data in, data out).
"""
from aif360.sklearn.preprocessing.reweighing import Reweighing, ReweighingMeta
from aif360.sklearn.preprocessing.fairadapt import FairAdapt
from aif360.sklearn.preprocessing.learning_fair_representations import LearnedFairRepresentations

__all__ = [
    'Reweighing', 'ReweighingMeta', 'FairAdapt', 'LearnedFairRepresentations'
]



from functools import partial
import numpy as np
import scipy.spatial.distance as scdist

from aif360.datasets import StructuredDataset
from aif360.metrics import DatasetMetric, utils


class SampleDistortionMetric(DatasetMetric):
    """Class for computing metrics based on two StructuredDatasets."""

    def __init__(self, dataset, distorted_dataset, unprivileged_groups=None,
                 privileged_groups=None):
        """
        Args:
            dataset (StructuredDataset): A StructuredDataset.
            distorted_dataset (StructuredDataset): A StructuredDataset.
            privileged_groups (list(dict)): Privileged groups. Format is a list
                of `dicts` where the keys are `protected_attribute_names` and
                the values are values in `protected_attributes`. Each `dict`
                element describes a single group. See examples for more details.
            unprivileged_groups (list(dict)): Unprivileged groups in the same
                format as `privileged_groups`.

        Raises:
            TypeError: `dataset` and `distorted_dataset` must be
                :obj:`~aif360.datasets.StructuredDataset` types.
        """
        # sets self.dataset, self.unprivileged_groups, self.privileged_groups
        super(SampleDistortionMetric, self).__init__(dataset,
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups)

        if isinstance(distorted_dataset, StructuredDataset):
            self.distorted_dataset = distorted_dataset
        else:
            raise TypeError("'distorted_dataset' should be a StructuredDataset")

        with dataset.temporarily_ignore('features', 'labels', 'scores'):
            if dataset != distorted_dataset:
                raise ValueError("The two datasets may differ in features and "
                                 "labels/scores only.")

    def total(self, dist, privileged):
        distance, weights = dist(privileged=privileged, returned=True)
        return np.sum(distance * weights)

    def average(self, dist, privileged):
        distance, weights = dist(privileged=privileged, returned=True)
        return np.average(distance, weights=weights)

    def maximum(self, dist, privileged):
        return np.max(dist(privileged=privileged))

    def euclidean_distance(self, privileged=None, returned=False):
        """Compute the average Euclidean distance between the samples from the
        two datasets.
        """
        condition = self._to_condition(privileged)
        distance, mask = utils.compute_distance(self.dataset.features,
            self.distorted_dataset.features, self.dataset.protected_attributes,
            self.dataset.protected_attribute_names, dist_fun=scdist.euclidean,
            condition=condition)
        if returned:
            return distance, self.dataset.instance_weights[mask]
        return distance

    def manhattan_distance(self, privileged=None, returned=False):
        """Compute the average Manhattan distance between the samples from the
        two datasets.
        """
        condition = self._to_condition(privileged)
        distance, mask = utils.compute_distance(self.dataset.features,
            self.distorted_dataset.features, self.dataset.protected_attributes,
            self.dataset.protected_attribute_names, dist_fun=scdist.cityblock,
            condition=condition)
        if returned:
            return distance, self.dataset.instance_weights[mask]
        return distance

    def mahalanobis_distance(self, privileged=None, returned=False):
        """Compute the average Mahalanobis distance between the samples from the
        two datasets.
        """
        condition = self._to_condition(privileged)
        X_orig = self.dataset.features
        X_distort = self.distorted_dataset.features
        dist_fun = partial(scdist.mahalanobis,
            VI=np.linalg.inv(np.cov(np.vstack([X_orig, X_distort]).T)).T)
        distance, mask = utils.compute_distance(X_orig, X_distort,
            self.dataset.protected_attributes,
            self.dataset.protected_attribute_names, dist_fun=dist_fun,
            condition=condition)
        if returned:
            return distance, self.dataset.instance_weights[mask]
        return distance

    def total_euclidean_distance(self, privileged=None):
        return self.total(self.euclidean_distance, privileged=privileged)

    def total_manhattan_distance(self, privileged=None):
        return self.total(self.manhattan_distance, privileged=privileged)

    def total_mahalanobis_distance(self, privileged=None):
        return self.total(self.mahalanobis_distance, privileged=privileged)

    def average_euclidean_distance(self, privileged=None):
        return self.average(self.euclidean_distance, privileged=privileged)

    def average_manhattan_distance(self, privileged=None):
        return self.average(self.manhattan_distance, privileged=privileged)

    def average_mahalanobis_distance(self, privileged=None):
        return self.average(self.mahalanobis_distance, privileged=privileged)

    def maximum_euclidean_distance(self, privileged=None):
        return self.maximum(self.euclidean_distance, privileged=privileged)

    def maximum_manhattan_distance(self, privileged=None):
        return self.maximum(self.manhattan_distance, privileged=privileged)

    def maximum_mahalanobis_distance(self, privileged=None):
        return self.maximum(self.mahalanobis_distance, privileged=privileged)

    def mean_euclidean_distance_difference(self, privileged=None):
        """Difference of the averages."""
        return self.difference(
            self.average(self.euclidean_distance, privileged=privileged))

    def mean_manhattan_distance_difference(self, privileged=None):
        """Difference of the averages."""
        return self.difference(
            self.average(self.manhattan_distance, privileged=privileged))

    def mean_mahalanobis_distance_difference(self, privileged=None):
        """Difference of the averages."""
        return self.difference(
            self.average(self.mahalanobis_distance, privileged=privileged))

    def mean_euclidean_distance_ratio(self, privileged=None):
        """Ratio of the averages."""
        return self.ratio(
            self.average(self.euclidean_distance, privileged=privileged))

    def mean_manhattan_distance_ratio(self, privileged=None):
        """Ratio of the averages."""
        return self.ratio(
            self.average(self.manhattan_distance, privileged=privileged))

    def mean_mahalanobis_distance_ratio(self, privileged=None):
        """Ratio of the averages."""
        return self.ratio(
            self.average(self.mahalanobis_distance, privileged=privileged))



# Adapted from:
# http://www.christianmoscardi.com/blog/2016/01/20/jupyter-testing.html

import os
import subprocess
import tempfile

import nbformat

def notebook_run(path):
    """Execute a notebook via nbconvert and collect output.
    Reset cwd after execution. 
       :returns (parsed nb object, execution errors)
    """
    old_cwd = os.getcwd()

    dirname, __ = os.path.split(path)
    os.chdir(dirname)

    kername = "python3"

    with tempfile.NamedTemporaryFile(suffix=".ipynb") as fout:
        args = ["jupyter", "nbconvert", "--to", "notebook", "--execute",
                "--ExecutePreprocessor.timeout=600",
                "--ExecutePreprocessor.allow_errors=True",
                "--ExecutePreprocessor.kernel_name={}".format(kername),
                "--output", fout.name, path]

        subprocess.check_call(args)

        fout.seek(0)
        nb = nbformat.read(fout, nbformat.current_nbformat)

    errors = [output for cell in nb.cells if "outputs" in cell
                     for output in cell["outputs"]
                     if output.output_type == "error"]
    
    os.chdir(old_cwd)

    return nb, errors



import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from aif360.sklearn.datasets import fetch_adult
from aif360.sklearn.metrics import statistical_parity_difference
from aif360.sklearn.preprocessing import fairadapt


X, y, sample_weight = fetch_adult(dropcols=['education', 'capital-gain',
                                            'capital-loss', 'relationship'])
X = X[0:5000]
y = y[0:5000]
(X_train, X_test,
 y_train, y_test) = train_test_split(X, y, train_size=0.8, random_state=1234567)

def test_fairadapt_adult():
    """Test that FairAdapt works when applied to Adult dataset."""
    train_df = pd.concat([X_train, y_train], axis=1)
    adj_mat = pd.DataFrame(
        np.zeros((len(train_df.columns), len(train_df.columns)), dtype=int),
        index=train_df.columns.values,
        columns=train_df.columns.values
    )

    # Construct the adjacency matrix of the causal graph
    adj_mat.loc[
        ["sex", "age", "native-country"],
        ["marital-status", "education-num", "workclass", "hours-per-week",
         "occupation", "annual-income"]
    ] = 1
    adj_mat.loc[
        "marital-status",
        ["education-num", "workclass", "hours-per-week", "occupation",
         "annual-income"]
    ] = 1
    adj_mat.loc[
        "education-num",
        ["workclass", "hours-per-week", "occupation", "annual-income"]
    ] = 1
    adj_mat.loc[
        ["workclass", "hours-per-week", "occupation"],
        "annual-income"
    ] = 1

    FA = fairadapt.FairAdapt(prot_attr="sex", adj_mat=adj_mat)
    Xf_train, yf_train, Xf_test = FA.fit_transform(X_train, y_train, X_test)

    # gap before adaptation
    gap = statistical_parity_difference(y_train, prot_attr="sex",
                                        priv_group="Male", pos_label=">50K")

    # gap after adaptation
    fair_gap = statistical_parity_difference(y_train, yf_train, prot_attr="sex",
                                             priv_group="Male", pos_label=">50K")

    assert isinstance(Xf_train, pd.DataFrame)
    assert isinstance(Xf_test, pd.DataFrame)
    assert isinstance(yf_train, pd.Series)
    assert all(Xf_train[FA.prot_attr] == "Female")
    assert abs(fair_gap) <= abs(gap) # assert that discrimination was reduced



# Copyright 2019 Seth V. Neel, Michael J. Kearns, Aaron L. Roth, Zhiwei Steven Wu
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


class ClassifierHistory(object):
    """Class ClassifierHistory storing lists of classifiers at each iteration and predictions."""
    def __init__(self):
        super(ClassifierHistory, self).__init__()
        self.classifiers = []
        self.predictions = []

    def append_classifier(self, classifier):
        self.classifiers.append(classifier)

    def get_most_recent(self):
        return self.classifiers[-1]


import sklearn.preprocessing
import numpy as np

from aif360.algorithms import Transformer


class LimeEncoder(Transformer):
    """Tranformer for converting aif360 dataset to LIME dataset and vice versa.

    (LIME - Local Interpretable Model-Agnostic Explanations) [2]_

    See for details/usage:
    https://github.com/marcotcr/lime

    References:
        .. [2] M.T. Ribeiro, S. Singh, and C. Guestrin, '"Why should I trust
           you?" Explaining the predictions of any classifier.'
           https://arxiv.org/pdf/1602.04938v1.pdf
    """

    def __init__(self):
        super(LimeEncoder, self).__init__()

    def fit(self, dataset):
        """Take an aif360 dataset and save all relevant metadata as well as
        mappings needed to transform/inverse_transform the data between aif360
        and lime.

        Args:
            dataset (BinaryLabelDataset): aif360 dataset

        Returns:
            LimeEncoder: Returns self.
        """
        self.s_feature_names_with_one_hot_encoding = dataset.feature_names
        df, df_dict = dataset.convert_to_dataframe(de_dummy_code=True)

        dfc = df.drop(dataset.label_names[0], axis=1)  # remove label (class) column

        self.s_feature_names = list(dfc.columns)       # create list of feature names
        self.s_data = dfc.values                       # create array of feature values

        # since categorical features are 1-hot-encoded and their names changed,
        # the set diff gives us the list of categorical features as non-
        # categorical feature names are not changed
        self.s_categorical_features = list(set(self.s_feature_names)
                                         - set(self.s_feature_names_with_one_hot_encoding))

        self.s_protected_attribute_names = dataset.protected_attribute_names

        # add protected attribute names to the list of categorical features
        self.s_categorical_features = self.s_categorical_features \
                                    + self.s_protected_attribute_names

        self.s_labels = df[dataset.label_names[0]]  # create labels

        # following 3 lines are not really needed
        # using to create s_class_names..can do so manually as well ...array([ 0.,  1.])
        s_le = sklearn.preprocessing.LabelEncoder()
        s_le.fit(self.s_labels)
        # self.s_labels = s_le.transform(self.s_labels)
        self.s_class_names = s_le.classes_

        # convert s_categorical_features to a list of array indexes in
        # s_feature_names corresponding to categorical features
        # (NOTE - does not included protected attributes)
        self.s_categorical_features = [self.s_feature_names.index(x)
                                       for x in self.s_categorical_features]

        # map all the categorical features to numerical values and store the
        # mappings in s_categorical_names
        self.s_categorical_names = {}
        for feature in self.s_categorical_features:
            self.le = sklearn.preprocessing.LabelEncoder()
            self.le.fit(self.s_data[:, feature])
            #self.s_data[:, feature] = le.transform(self.s_data[:, feature])
            self.s_categorical_names[feature] = self.le.classes_

        return self

    def transform(self, aif360data):
        """Take aif360 data array and return data array that is lime encoded
        (numeric array in which categorical features are NOT one-hot-encoded).

        Args:
            aif360data (np.ndarray): Dataset features

        Returns:
            np.ndarray: LIME dataset features
        """
        tgtNumRows = aif360data.shape[0]
        tgtNumcolumns = len(self.s_feature_names)
        limedata = np.zeros(shape=(tgtNumRows, tgtNumcolumns))

        # non_categorical_features = list(set(self.s_feature_names) & set(self.s_feature_names_with_one_hot_encoding))
        for rw in range(limedata.shape[0]):
            for ind, feature in enumerate(self.s_feature_names):
                if ind in self.s_categorical_features:
                    # tranform the value since categorical feature except if it
                    # is also a protected attribute
                    if feature in self.s_protected_attribute_names:
                        # just copy the value as is
                        limedata[rw, ind] = aif360data[rw, self.s_feature_names_with_one_hot_encoding.index(feature)]
                    else:
                        possible_feature_values = self.s_categorical_names[ind]
                        for indc in range(len(possible_feature_values)):
                            cval = possible_feature_values[indc]
                            colName = feature + "=" + cval
                            if (aif360data[rw][self.s_feature_names_with_one_hot_encoding.index(colName)] == 1.0):
                                limedata[rw][ind] = indc
                else:
                    # just copy the value as is
                    limedata[rw, ind] = aif360data[rw, self.s_feature_names_with_one_hot_encoding.index(feature)]

        return limedata

    def inverse_transform(self, limedata):
        """Take data array that is lime encoded (that is, lime-compatible data
        created by this class from a given aif360 dataset) and return data array
        consistent with the original aif360 dataset.

        Args:
            limedata (np.ndarray): Dataset features

        Returns:
            np.ndarray: aif360 dataset features
        """
        tgtNumRows = limedata.shape[0]
        tgtNumcolumns = len(self.s_feature_names_with_one_hot_encoding)
        aif360data = np.zeros(shape=(tgtNumRows, tgtNumcolumns))

        for rw in range(aif360data.shape[0]):
            for ind, feature in enumerate(self.s_feature_names):
                # s_categorical_features has list of indexes into
                # s_feature_names for categorical features
                if ind in self.s_categorical_features:
                    if feature in self.s_protected_attribute_names:
                        # just copy the value as is
                        aif360data[rw, self.s_feature_names_with_one_hot_encoding.index(feature)] = limedata[rw, ind]
                    else:
                        # s_categorical_names[ind] has mapping of categorical to
                        # numerical values i.e. limedata[rw, ind] is index of
                        # this array. value is string val
                        new_feature = feature + '=' + self.s_categorical_names[ind][int(limedata[rw, ind])]
                        # categorical feature:
                        aif360data[rw, self.s_feature_names_with_one_hot_encoding.index(new_feature)] = 1.0
                else: # just copy value
                    aif360data[rw, self.s_feature_names_with_one_hot_encoding.index(feature)] = limedata[rw, ind]

        return aif360data



import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LogisticRegression

from aif360.datasets import AdultDataset
from aif360.metrics import BinaryLabelDatasetMetric, ClassificationMetric

ad = AdultDataset(protected_attribute_names=['race', 'sex', 'native-country'],
                  privileged_classes=[['White'], ['Male'], ['United-States']],
                  categorical_features=['workclass', 'education',
                          'marital-status', 'occupation', 'relationship'],
                  custom_preprocessing=lambda df: df.fillna('Unknown'))
adult_test, adult_train = ad.split([16281], shuffle=False)

scaler = StandardScaler()
X = scaler.fit_transform(adult_train.features)
test_X = scaler.transform(adult_test.features)
clf = LogisticRegression(C=1.0, random_state=0, solver='liblinear')

adult_pred = adult_test.copy()
adult_pred.labels = clf.fit(X, adult_train.labels.ravel()).predict(test_X)

dataset_metric = BinaryLabelDatasetMetric(adult_test)
classifier_metric = BinaryLabelDatasetMetric(adult_pred)

def test_epsilon_dataset_binary_groups():
    eps_data = dataset_metric.smoothed_empirical_differential_fairness()
    assert eps_data == 1.53679014653623  # verified with reference implementation

def test_epsilon_classifier_binary_groups():
    eps_clf = classifier_metric.smoothed_empirical_differential_fairness()
    assert eps_clf == 1.6434003346776307  # verified with reference implementation

def test_bias_amplification_binary_groups():
    metric = ClassificationMetric(adult_test, adult_pred)
    bias_amp = metric.differential_fairness_bias_amplification()
    eps_data = dataset_metric.smoothed_empirical_differential_fairness()
    eps_clf = classifier_metric.smoothed_empirical_differential_fairness()
    assert bias_amp == (eps_clf - eps_data)

def test_epsilon_all_groups():
    def custom_preprocessing(df):
        # slight workaround for non-binary protected attribute
        # feature should be categorical but protected attribute should be numerical
        mapping = {'Black': 0, 'White': 1, 'Asian-Pac-Islander': 2,
                   'Amer-Indian-Eskimo': 3, 'Other': 4}
        df['race-num'] = df.race.map(mapping)
        return df.fillna('Unknown')

    nonbinary_ad = AdultDataset(
            protected_attribute_names=['sex', 'native-country', 'race-num'],
            privileged_classes=[['Male'], ['United-States'], [1]],
            categorical_features=['workclass', 'education', 'marital-status',
                                  'occupation', 'relationship', 'race'],
            custom_preprocessing=custom_preprocessing)
    # drop redundant race feature (not relevant to this test)
    index = nonbinary_ad.feature_names.index('race-num')
    nonbinary_ad.features = np.delete(nonbinary_ad.features, index, axis=1)
    nonbinary_ad.feature_names = np.delete(nonbinary_ad.feature_names, index)

    nonbinary_test, _ = nonbinary_ad.split([16281], shuffle=False)
    dataset_metric = BinaryLabelDatasetMetric(nonbinary_test)
    eps_data = dataset_metric.smoothed_empirical_differential_fairness()
    assert eps_data == 2.063813731996515  # verified with reference implementation



from logging import warning

import numpy as np
import pandas as pd

from aif360.datasets import BinaryLabelDataset


class StandardDataset(BinaryLabelDataset):
    """Base class for every :obj:`BinaryLabelDataset` provided out of the box by
    aif360.

    It is not strictly necessary to inherit this class when adding custom
    datasets but it may be useful.

    This class is very loosely based on code from
    https://github.com/algofairness/fairness-comparison.
    """

    def __init__(self, df, label_name, favorable_classes,
                 protected_attribute_names, privileged_classes,
                 instance_weights_name='', scores_name='',
                 categorical_features=[], features_to_keep=[],
                 features_to_drop=[], na_values=[], custom_preprocessing=None,
                 metadata=None):
        """
        Subclasses of StandardDataset should perform the following before
        calling `super().__init__`:

            1. Load the dataframe from a raw file.

        Then, this class will go through a standard preprocessing routine which:

            2. (optional) Performs some dataset-specific preprocessing (e.g.
               renaming columns/values, handling missing data).

            3. Drops unrequested columns (see `features_to_keep` and
               `features_to_drop` for details).

            4. Drops rows with NA values.

            5. Creates a one-hot encoding of the categorical variables.

            6. Maps protected attributes to binary privileged/unprivileged
               values (1/0).

            7. Maps labels to binary favorable/unfavorable labels (1/0).

        Args:
            df (pandas.DataFrame): DataFrame on which to perform standard
                processing.
            label_name: Name of the label column in `df`.
            favorable_classes (list or function): Label values which are
                considered favorable or a boolean function which returns `True`
                if favorable. All others are unfavorable. Label values are
                mapped to 1 (favorable) and 0 (unfavorable) if they are not
                already binary and numerical.
            protected_attribute_names (list): List of names corresponding to
                protected attribute columns in `df`.
            privileged_classes (list(list or function)): Each element is
                a list of values which are considered privileged or a boolean
                function which return `True` if privileged for the corresponding
                column in `protected_attribute_names`. All others are
                unprivileged. Values are mapped to 1 (privileged) and 0
                (unprivileged) if they are not already numerical.
            instance_weights_name (optional): Name of the instance weights
                column in `df`.
            categorical_features (optional, list): List of column names in the
                DataFrame which are to be expanded into one-hot vectors.
            features_to_keep (optional, list): Column names to keep. All others
                are dropped except those present in `protected_attribute_names`,
                `categorical_features`, `label_name` or `instance_weights_name`.
                Defaults to all columns if not provided.
            features_to_drop (optional, list): Column names to drop. *Note: this
                overrides* `features_to_keep`.
            na_values (optional): Additional strings to recognize as NA. See
                :func:`pandas.read_csv` for details.
            custom_preprocessing (function): A function object which
                acts on and returns a DataFrame (f: DataFrame -> DataFrame). If
                `None`, no extra preprocessing is applied.
            metadata (optional): Additional metadata to append.
        """
        # 2. Perform dataset-specific preprocessing
        if custom_preprocessing:
            df = custom_preprocessing(df)

        # 3. Drop unrequested columns
        features_to_keep = features_to_keep or df.columns.tolist()
        keep = (set(features_to_keep) | set(protected_attribute_names)
              | set(categorical_features) | set([label_name]))
        if instance_weights_name:
            keep |= set([instance_weights_name])
        df = df[sorted(keep - set(features_to_drop), key=df.columns.get_loc)]
        categorical_features = sorted(set(categorical_features) - set(features_to_drop), key=df.columns.get_loc)

        # 4. Remove any rows that have missing data.
        dropped = df.dropna()
        count = df.shape[0] - dropped.shape[0]
        if count > 0:
            warning("Missing Data: {} rows removed from {}.".format(count,
                    type(self).__name__))
        df = dropped

        # 5. Create a one-hot encoding of the categorical variables.
        df = pd.get_dummies(df, columns=categorical_features, prefix_sep='=')

        # 6. Map protected attributes to privileged/unprivileged
        privileged_protected_attributes = []
        unprivileged_protected_attributes = []
        for attr, vals in zip(protected_attribute_names, privileged_classes):
            privileged_values = [1.]
            unprivileged_values = [0.]
            if callable(vals):
                df[attr] = df[attr].apply(vals)
            elif np.issubdtype(df[attr].dtype, np.number):
                # this attribute is numeric; no remapping needed
                privileged_values = vals
                unprivileged_values = list(set(df[attr]).difference(vals))
            else:
                # find all instances which match any of the attribute values
                priv = np.logical_or.reduce(np.equal.outer(vals, df[attr].to_numpy()))
                df.loc[priv, attr] = privileged_values[0]
                df.loc[~priv, attr] = unprivileged_values[0]

            privileged_protected_attributes.append(
                np.array(privileged_values, dtype=np.float64))
            unprivileged_protected_attributes.append(
                np.array(unprivileged_values, dtype=np.float64))

        # 7. Make labels binary
        favorable_label = 1.
        unfavorable_label = 0.
        if callable(favorable_classes):
            df[label_name] = df[label_name].apply(favorable_classes)
        elif np.issubdtype(df[label_name], np.number) and len(set(df[label_name])) == 2:
            # labels are already binary; don't change them
            favorable_label = favorable_classes[0]
            unfavorable_label = set(df[label_name]).difference(favorable_classes).pop()
        else:
            # find all instances which match any of the favorable classes
            pos = np.logical_or.reduce(np.equal.outer(favorable_classes, 
                                                      df[label_name].to_numpy()))
            df.loc[pos, label_name] = favorable_label
            df.loc[~pos, label_name] = unfavorable_label

        super(StandardDataset, self).__init__(df=df, label_names=[label_name],
            protected_attribute_names=protected_attribute_names,
            privileged_protected_attributes=privileged_protected_attributes,
            unprivileged_protected_attributes=unprivileged_protected_attributes,
            instance_weights_name=instance_weights_name,
            scores_names=[scores_name] if scores_name else [],
            favorable_label=favorable_label,
            unfavorable_label=unfavorable_label, metadata=metadata)



import numpy as np
import pytest
import sklearn
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split

from aif360.algorithms.postprocessing import CalibratedEqOddsPostprocessing
from aif360.sklearn.postprocessing import CalibratedEqualizedOdds, PostProcessingMeta


def test_calib_eq_odds_sex_weighted(old_adult, new_adult):
    """Test that the old and new CalibratedEqualizedOdds produce the same mix
    rates.
    """
    X, y, sample_weight = new_adult
    logreg = LogisticRegression(solver='lbfgs', max_iter=500)
    y_pred = logreg.fit(X, y, sample_weight=sample_weight).predict_proba(X)
    adult_pred = old_adult.copy()
    adult_pred.scores = y_pred[:, 1]
    orig_cal_eq_odds = CalibratedEqOddsPostprocessing(
            unprivileged_groups=[{'sex': 0}], privileged_groups=[{'sex': 1}])
    orig_cal_eq_odds.fit(old_adult, adult_pred)
    cal_eq_odds = CalibratedEqualizedOdds('sex')
    cal_eq_odds.fit(y_pred, y, sample_weight=sample_weight)

    assert np.isclose(orig_cal_eq_odds.priv_mix_rate, cal_eq_odds.mix_rates_[1])
    assert np.isclose(orig_cal_eq_odds.unpriv_mix_rate, cal_eq_odds.mix_rates_[0])

def test_postprocessingmeta_fnr(old_adult, new_adult):
    """Test that the old and new CalibratedEqualizedOdds produce the same
    probability predictions.

    This tests the whole "pipeline": splitting the data the same way, training a
    LogisticRegression classifier, and training the post-processor the same way.
    """
    X, y, sample_weight = new_adult
    adult_train, adult_test = old_adult.split([0.9], shuffle=False)
    X_tr, X_te, y_tr, _, sw_tr, _ = train_test_split(X, y, sample_weight,
                train_size=0.9, shuffle=False)

    assert np.all(adult_train.features == X_tr)
    assert np.all(adult_test.features == X_te)
    assert np.all(adult_train.labels.ravel() == y_tr)

    adult_est, adult_post = adult_train.split([0.75], shuffle=False)

    logreg = LogisticRegression(solver='lbfgs', max_iter=500)
    logreg.fit(adult_est.features, adult_est.labels.ravel(),
               sample_weight=adult_est.instance_weights)
    probas_pred = logreg.predict_proba(adult_post.features)[:, 1]

    adult_pred = adult_post.copy()
    adult_pred.scores = probas_pred

    orig_cal_eq_odds = CalibratedEqOddsPostprocessing(
            unprivileged_groups=[{'sex': 0}], privileged_groups=[{'sex': 1}],
            cost_constraint='fnr', seed=0)
    orig_cal_eq_odds.fit(adult_post, adult_pred)

    cal_eq_odds = PostProcessingMeta(estimator=logreg,
            postprocessor=CalibratedEqualizedOdds('sex', cost_constraint='fnr',
                                                  random_state=0),
            shuffle=False)
    cal_eq_odds.fit(X_tr, y_tr, sample_weight=sw_tr)

    assert np.allclose(logreg.coef_, cal_eq_odds.estimator_.coef_)

    assert np.allclose([orig_cal_eq_odds.unpriv_mix_rate,
                        orig_cal_eq_odds.priv_mix_rate],
                       cal_eq_odds.postprocessor_.mix_rates_)

    adult_test_pred = adult_test.copy()
    adult_test_pred.scores = logreg.predict_proba(adult_test.features)[:, 1]
    adult_test_pred = orig_cal_eq_odds.predict(adult_test_pred)

    y_test_pred = cal_eq_odds.predict_proba(X_te)

    assert np.allclose(adult_test_pred.scores, y_test_pred[:, 1])

def test_calib_eq_odds_prefit(new_adult):
    """Test the 'prefit' option in PostProcessingMeta."""
    X, y, sample_weight = new_adult

    logreg = LogisticRegression(solver='lbfgs', max_iter=500)
    cal_eq_odds = CalibratedEqualizedOdds('sex')
    pp = PostProcessingMeta(logreg, cal_eq_odds, prefit=True)
    with pytest.raises(sklearn.exceptions.NotFittedError):
        pp.fit(X, y, sample_weight=sample_weight)

    pp = PostProcessingMeta(logreg, cal_eq_odds, random_state=1234)
    pp.fit(X, y, sample_weight=sample_weight)

    _, X_pp, _, y_pp, _, sw_pp = train_test_split(X, y, sample_weight, random_state=1234)
    pp_prefit = PostProcessingMeta(pp.estimator_, cal_eq_odds, prefit=True)
    pp_prefit.fit(X_pp, y_pp, sample_weight=sw_pp)

    assert np.allclose(pp.postprocessor_.mix_rates_, pp_prefit.postprocessor_.mix_rates_)



import warnings

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import GridSearchCV
from sklearn.utils.validation import check_is_fitted

from aif360.sklearn.metrics import statistical_parity_difference
from aif360.sklearn.metrics import average_odds_error
from aif360.sklearn.metrics import equal_opportunity_difference
from aif360.sklearn.metrics import disparate_impact_ratio
from aif360.sklearn.metrics import make_scorer
from aif360.sklearn.utils import check_groups


class RejectOptionClassifier(BaseEstimator, ClassifierMixin):
    """Reject option based classification (ROC) post-processor.

    Reject option classification is a post-processing technique that gives
    favorable outcomes to unprivileged groups and unfavorable outcomes to
    privileged groups in a confidence band around the decision boundary with the
    highest uncertainty [#kamiran12]_.

    Note:
        A :class:`~sklearn.pipeline.Pipeline` expects a single estimation step
        but this class requires an estimator's predictions as input. See
        :class:`PostProcessingMeta` for a workaround.

    See also:
        :class:`PostProcessingMeta`, :class:`RejectOptionClassifierCV`

    References:
        .. [#kamiran12] `F. Kamiran, A. Karim, and X. Zhang, "Decision Theory
           for Discrimination-Aware Classification," IEEE International
           Conference on Data Mining, 2012.
           <https://ieeexplore.ieee.org/abstract/document/6413831>`_

    Attributes:
        prot_attr_ (str or list(str)): Protected attribute(s) used for post-
            processing.
        groups_ (array, shape (2,)): A list of group labels known to the
            classifier. Note: this algorithm require a binary division of the
            data.
        classes_ (array, shape (num_classes,)): A list of class labels known to
            the classifier. Note: this algorithm treats all non-positive
            outcomes as negative (binary classification only).
        pos_label_ (scalar): The label of the positive class.
        priv_group_ (scalar): The label of the privileged group.

    Examples:
        RejectOptionClassifier can be easily paired with GridSearchCV to
        find the best threshold and margin with respect to a fairness measure:

        >>> from sklearn.model_selection import GridSearchCV
        >>> roc = RejectOptionClassifier()
        >>> param = [{'threshold': [t],
                      'margin': np.arange(0.05, min(t, 1-t)+0.025, 0.05)}
        ...          for t in np.arange(0.05, 1., 0.05)]
        >>> stat_par = make_scorer(statistical_parity_difference)
        >>> scoring = {'bal_acc': 'balanced_accuracy', 'stat_par': stat_par}
        >>> def refit(cv_res):
        ...     return np.ma.array(cv_res['mean_test_bal_acc'],
        ...             mask=cv_res['mean_test_stat_par'] < -0.1).argmax()
        ...
        >>> grid = GridSearchCV(roc, param, scoring=scoring, refit=refit)

        Or, alternatively, this can be done in one step with
        RejectOptionClassifierCV:

        >>> grid = RejectOptionClassifierCV(scoring='statistical_parity')

    """
    def __init__(self, prot_attr=None, threshold=0.5, margin=0.1):
        """
        Args:
            prot_attr (single label or list-like, optional): Protected
                attribute(s) to use in the post-processing. If more than one
                attribute, all combinations of values (intersections) are
                considered. Default is ``None`` meaning all protected attributes
                from the dataset are used. Note: This algorithm requires there
                be exactly 2 groups (privileged and unprivileged).
            threshold (scalar): Classification threshold. Probability estimates
                greater than this value are considered positive. Must be between
                0 and 1.
            margin (scalar): Half width of the critical region. Estimates within
                the critical region are "rejected" and assigned according to
                their group. Must be between 0 and min(threshold, 1-threshold).
        """
        self.prot_attr = prot_attr
        self.threshold = threshold
        self.margin = margin

    def _more_tags(self):
        return {'requires_proba': True}

    def fit(self, X, y, labels=None, pos_label=1, priv_group=1,
            sample_weight=None):
        """This is essentially a no-op; it simply validates the inputs and
        stores them for predict.

        Args:
            X (array-like): Ignored.
            y (array-like): Ground-truth (correct) target values. Note: one of X
                or y must contain protected attribute information.
            labels (list, optional): The ordered set of labels values. Must
                match the order of columns in X if provided. By default,
                all labels in y are used in sorted order.
            pos_label (scalar, optional): The label of the positive class.
            priv_group (scalar, optional): The label of the privileged group.
            sample_weight (array-like, optional): Ignored.

        Returns:
            self
        """
        try:
            groups, self.prot_attr_ = check_groups(X, self.prot_attr,
                                                   ensure_binary=True)
        except TypeError:
            groups, self.prot_attr_ = check_groups(y, self.prot_attr,
                                                   ensure_binary=True)
        self.classes_ = np.array(labels) if labels is not None else np.unique(y)
        self.groups_ = np.unique(groups)
        self.pos_label_ = pos_label
        self.priv_group_ = priv_group

        if len(self.classes_) != 2:
            raise ValueError('Only binary classification is supported.')

        if pos_label not in self.classes_:
            raise ValueError('pos_label={} is not in the set of labels. The '
                    'valid values are:\n{}'.format(pos_label, self.classes_))

        if priv_group not in self.groups_:
            raise ValueError('priv_group={} is not in the set of groups. The '
                    'valid values are:\n{}'.format(priv_group, self.groups_))

        if not 0.0 <= self.threshold <= 1.0:
            raise ValueError('threshold must be between 0.0 and 1.0, '
                             'threshold={}'.format(self.threshold))

        if not 0.0 <= self.margin <= min(self.threshold, 1 - self.threshold):
            raise ValueError('margin must be between 0.0 and {}, margin={}'
                             ''.format(min(self.threshold, 1 - self.threshold),
                                       self.margin))

        return self

    def predict_proba(self, X):
        """Probability estimates.

        The returned estimates for all classes are ordered by the label of
        classes.

        Args:
            X (pandas.DataFrame): Probability estimates of the targets as
                returned by a ``predict_proba()`` call or equivalent. Note: must
                include protected attributes in the index.
        Returns:
            numpy.ndarray: Returns the probability of the sample for each class
            in the model, where classes are ordered as they are in
            ``self.classes_``.
        """
        check_is_fitted(self, 'pos_label_')

        groups, _ = check_groups(X, self.prot_attr_)
        if len(self.classes_) != X.shape[1]:
            raise ValueError('X should contain one column per class. Got: {} '
                             'columns.'.format(X.shape[1]))

        pos_idx = np.nonzero(self.classes_ == self.pos_label_)[0][0]
        yt = X.iloc[:, pos_idx].to_numpy().copy()

        # indices of critical region around the classification boundary
        crit_above = (self.margin > yt-self.threshold) & (yt > self.threshold)
        crit_below = (-self.margin < yt-self.threshold) & (yt < self.threshold)

        # flip labels: priv + above -> below, unpriv + below -> above
        priv = (groups == self.priv_group_)
        flip = (priv & crit_above) | (~priv & crit_below)
        yt[flip] = 2*self.threshold - yt[flip]

        return np.c_[1 - yt, yt] if pos_idx == 1 else np.c_[yt, 1 - yt]

    def predict(self, X):
        """Predict class labels for the given scores.

        Args:
            X (pandas.DataFrame): Probability estimates of the targets as
                returned by a ``predict_proba()`` call or equivalent. Note: must
                include protected attributes in the index.

        Returns:
            numpy.ndarray: Predicted class label per sample.
        """
        scores = self.predict_proba(X)
        pos_idx = np.nonzero(self.classes_ == self.pos_label_)[0][0]
        y_pred = (scores[:, pos_idx] > self.threshold).astype(int)
        return self.classes_[y_pred if pos_idx == 1 else 1 - y_pred]

    def fit_predict(self, X, y=None, **fit_params):
        """Predict class labels for the given scores.

        In general, it is not necessary to fit and predict separately so this
        method may be used instead. For subsequent predicts, it may be easier
        to use the `predict` method, though.

        Args:
            X (pandas.DataFrame): Probability estimates of the targets as
                returned by a ``predict_proba()`` call or equivalent. Note: must
                include protected attributes in the index.
            y (array-like, optional): Ground-truth (correct) target values.
                Note: if not provided, `labels` must be provided in
                `**fit_params`. See `fit` for details.
            **fit_params: See `fit` for details.

        Returns:
            numpy.ndarray: Predicted class label per sample.
        """
        return self.fit(X, y, **fit_params).predict(X)


class RejectOptionClassifierCV(GridSearchCV):
    """Wrapper for running a grid search over threshold, margin combinations for
    a RejectOptionClassifier.

    Note:
        :class:`~sklearn.model_selection.GridSearchCV` does not currently
        support sample weights in scoring. This will work but throw a warning if
        `sample_weight` is provided.

    See also:
        :class:`RejectOptionClassifier`

    Examples:
        >>> import matplotlib.pyplot as plt
        >>> import pandas as pd
        >>> from sklearn.linear_model import LogisticRegression
        >>> from aif360.sklearn.datasets import fetch_german
        >>> from aif360.sklearn.postprocessing import RejectOptionClassifierCV
        >>> X, y = fetch_german(numeric_only=True)
        >>> lr = LogisticRegression(solver='lbfgs').fit(X, y)
        >>> roc = RejectOptionClassifierCV('sex', scoring='disparate_impact')
        >>> roc.fit(pd.DataFrame(lr.predict_proba(X), index=X.index), y)

        We can also achieve this more simply using a PostProcessingMeta
        estimator:

        >>> from aif360.sklearn.postprocessing import PostProcessingMeta
        >>> pp = PostProcessingMeta(lr, roc).fit(X, y)
    """
    def __init__(self, prot_attr=None, *, scoring, step=0.05, refit=True, **kwargs):
        """
        Args:
            prot_attr (single label or list-like, optional): Protected
                attribute(s) to use in the post-processing. If more than one
                attribute, all combinations of values (intersections) are
                considered. Default is ``None`` meaning all protected attributes
                from the dataset are used. Note: This algorithm requires there
                be exactly 2 groups (privileged and unprivileged).
            scoring ('statistical_parity', 'average_odds', 'equal_opportunity',
                'disparate_impact', or callable/dict): Fairness scorer to use to
                evaluate the predictions. If type is a `str`, constructs the
                corresponding scorer for that metric in addition to the default
                balanced accuracy. If type is callable (i.e., a scorer object),
                that will be used along with balanced accuracy. Finally, if an
                explicit dictionary is passed, this will be used as is.
            step (float): Step size for grid search. Will search every valid
                combination of threshold and margin that are multiples of this
                step size. See `param_grid` after fitting for the exact search
                space.
            refit (bool or callable, optional): Refit the estimator using the
                best parameters found. If `True` and not using a custom scoring
                function, this chooses the highest balanced accuracy given
                fairness score > -0.1 (or > 0.8 for disparate impact only).
                Alternatively, a custom refitting function may be passed. See
                :class:`~sklearn.model_selection.GridSearchCV` for details.
            **kwargs: See :class:`~sklearn.model_selection.GridSearchCV` for
                additional kwargs.
        """
        self.scoring = scoring
        self.refit = refit
        self.step = step
        self.prot_attr = prot_attr
        super().__init__(RejectOptionClassifier(), {}, scoring=scoring,
                         refit=refit, **kwargs)

    def _more_tags(self):
        return {'requires_proba': True}

    def fit(self, X, y, **fit_params):
        """Run fit with all sets of parameters.

        Args:
            X (pandas.DataFrame): Probability estimates of the targets as
                returned by a ``predict_proba()`` call or equivalent. Note: must
                include protected attributes in the index.
            y (pandas.Series): Ground-truth (correct) target values.
            **fit_params: Parameters passed to the ``fit()`` method.

        Returns:
            self
        """
        self.param_grid = []
        thresholds = np.arange(self.step, 1, self.step)
        # arange has numerical instabilities. this way guarantees margin <= threshold
        for i, t in enumerate(thresholds):
            n = min(i+1, len(thresholds)-i)
            margins = np.linspace(min(self.step, min(t, 1-t)), min(t, 1-t), n)
            self.param_grid.append({'prot_attr': [self.prot_attr],
                                    'threshold': [t], 'margin': margins})

        if fit_params.get('sample_weight', None) is not None:
            warnings.warn('sample_weight will be ignored when scoring.',
                          RuntimeWarning)

        if not isinstance(self.scoring, dict):
            # TODO: sample_weight scoring workaround
            self.scorer_name_ = self.scoring
            if self.scoring == 'statistical_parity':
                self.scorer_ = make_scorer(statistical_parity_difference,
                        prot_attr=self.prot_attr)
            elif self.scoring == 'average_odds':
                self.scorer_ = make_scorer(average_odds_error,
                        prot_attr=self.prot_attr)
            elif self.scoring == 'equal_opportunity':
                self.scorer_ = make_scorer(equal_opportunity_difference,
                        prot_attr=self.prot_attr)
            elif self.scoring == 'disparate_impact':
                self.scorer_ = make_scorer(disparate_impact_ratio, is_ratio=True,
                        prot_attr=self.prot_attr, zero_division=0)
            elif not callable(self.scoring):
                raise ValueError("scorer must be one of: 'statistical_parity', "
                    "'average_odds', 'equal_opportunity', 'disparate_impact' "
                    "or a callable function. Got:\n{}".format(self.scoring))
            else:
                self.scorer_name_ = 'fairness_metric'
                self.scorer_ = self.scoring

            self.scoring = {'bal_acc': 'balanced_accuracy',
                            self.scorer_name_: self.scorer_}

        if self.refit is True and self.scorer_name_ != 'fairness_metric':
            if self.scorer_name_ == 'disparate_impact':
                self.refit = lambda res: np.ma.array(res['mean_test_bal_acc'],
                    mask=res['mean_test_disparate_impact'] < 0.8).argmax()
            else:
                self.refit = lambda res: np.ma.array(res['mean_test_bal_acc'],
                    mask=res['mean_test_'+self.scorer_name_] < -0.1).argmax()

        class NoSplit:
            def split(self, X, y=None, groups=None):
                yield np.arange(len(X)), np.arange(len(X))

            def get_n_splits(self, X=None, y=None, groups=None):
                return 1

        self.cv = NoSplit()

        return super().fit(X, y, **fit_params)



import aif360.algorithms.inprocessing.gerryfair.auditor
import aif360.algorithms.inprocessing.gerryfair.classifier_history
import aif360.algorithms.inprocessing.gerryfair.clean
import aif360.algorithms.inprocessing.gerryfair.fairness_plots
import aif360.algorithms.inprocessing.gerryfair.heatmap
import aif360.algorithms.inprocessing.gerryfair.learner
import aif360.algorithms.inprocessing.gerryfair.reg_oracle_class
__name__ = "gerryfair"
__version__ = "0.1.0"



import numpy as np
from sklearn.neighbors import NearestNeighbors
from aif360.algorithms.inprocessing.gerryfair.auditor import Auditor
from aif360.datasets import BinaryLabelDataset
from aif360.datasets.multiclass_label_dataset import MulticlassLabelDataset
from aif360.metrics import DatasetMetric, utils
from aif360.algorithms.inprocessing.gerryfair.clean import *


class BinaryLabelDatasetMetric(DatasetMetric):
    """Class for computing metrics based on a single
    :obj:`~aif360.datasets.BinaryLabelDataset`.
    """

    def __init__(self, dataset, unprivileged_groups=None, privileged_groups=None):
        """
        Args:
            dataset (BinaryLabelDataset): A BinaryLabelDataset.
            privileged_groups (list(dict)): Privileged groups. Format is a list
                of `dicts` where the keys are `protected_attribute_names` and
                the values are values in `protected_attributes`. Each `dict`
                element describes a single group. See examples for more details.
            unprivileged_groups (list(dict)): Unprivileged groups in the same
                format as `privileged_groups`.

        Raises:
            TypeError: `dataset` must be a
                :obj:`~aif360.datasets.BinaryLabelDataset` type.
        """
        if not isinstance(dataset, BinaryLabelDataset) and not isinstance(dataset, MulticlassLabelDataset) :
            raise TypeError("'dataset' should be a BinaryLabelDataset or a MulticlassLabelDataset")

        # sets self.dataset, self.unprivileged_groups, self.privileged_groups
        super(BinaryLabelDatasetMetric, self).__init__(dataset,
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups)

        if isinstance(dataset, MulticlassLabelDataset):
            fav_label_value = 1.
            unfav_label_value = 0.

            self.dataset = self.dataset.copy()
            # Find all labels which match any of the favorable labels
            fav_idx = np.logical_or.reduce(np.equal.outer(self.dataset.favorable_label, self.dataset.labels))
            # Replace labels with corresponding values
            self.dataset.labels = np.where(fav_idx, fav_label_value, unfav_label_value)

            self.dataset.favorable_label = float(fav_label_value)
            self.dataset.unfavorable_label = float(unfav_label_value)

    def num_positives(self, privileged=None):
        r"""Compute the number of positives,
        :math:`P = \sum_{i=1}^n \mathbb{1}[y_i = 1]`,
        optionally conditioned on protected attributes.

        Args:
            privileged (bool, optional): Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.

        Raises:
            AttributeError: `privileged_groups` or `unprivileged_groups` must be
                must be provided at initialization to condition on them.
        """
        condition = self._to_condition(privileged)
        return utils.compute_num_pos_neg(self.dataset.protected_attributes,
            self.dataset.labels, self.dataset.instance_weights,
            self.dataset.protected_attribute_names,
            self.dataset.favorable_label, condition=condition)

    def num_negatives(self, privileged=None):
        r"""Compute the number of negatives,
        :math:`N = \sum_{i=1}^n \mathbb{1}[y_i = 0]`, optionally conditioned on
        protected attributes.

        Args:
            privileged (bool, optional): Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.

        Raises:
            AttributeError: `privileged_groups` or `unprivileged_groups` must be
                must be provided at initialization to condition on them.
        """
        condition = self._to_condition(privileged)
        return utils.compute_num_pos_neg(self.dataset.protected_attributes,
            self.dataset.labels, self.dataset.instance_weights,
            self.dataset.protected_attribute_names,
            self.dataset.unfavorable_label, condition=condition)

    def base_rate(self, privileged=None):
        """Compute the base rate, :math:`Pr(Y = 1) = P/(P+N)`, optionally
        conditioned on protected attributes.

        Args:
            privileged (bool, optional): Boolean prescribing whether to
                condition this metric on the `privileged_groups`, if `True`, or
                the `unprivileged_groups`, if `False`. Defaults to `None`
                meaning this metric is computed over the entire dataset.
        Returns:
            float: Base rate (optionally conditioned).
        """
        return (self.num_positives(privileged=privileged)
              / self.num_instances(privileged=privileged))

    def disparate_impact(self):
        r"""
        .. math::
           \frac{Pr(Y = 1 | D = \text{unprivileged})}
           {Pr(Y = 1 | D = \text{privileged})}
        """
        return self.ratio(self.base_rate)

    def statistical_parity_difference(self):
        r"""
        .. math::
           Pr(Y = 1 | D = \text{unprivileged})
           - Pr(Y = 1 | D = \text{privileged})
        """
        return self.difference(self.base_rate)

    def consistency(self, n_neighbors=5):
        r"""Individual fairness metric from [1]_ that measures how similar the
        labels are for similar instances.

        .. math::
           1 - \frac{1}{n}\sum_{i=1}^n |\hat{y}_i -
           \frac{1}{\text{n_neighbors}} \sum_{j\in\mathcal{N}_{\text{n_neighbors}}(x_i)} \hat{y}_j|

        Args:
            n_neighbors (int, optional): Number of neighbors for the knn
                computation.

        References:
            .. [1] R. Zemel, Y. Wu, K. Swersky, T. Pitassi, and C. Dwork,
               "Learning Fair Representations,"
               International Conference on Machine Learning, 2013.
        """

        X = self.dataset.features
        num_samples = X.shape[0]
        y = self.dataset.labels

        # learn a KNN on the features
        nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm='ball_tree')
        nbrs.fit(X)
        _, indices = nbrs.kneighbors(X)

        # compute consistency score
        consistency = 0.0
        for i in range(num_samples):
            consistency += np.abs(y[i] - np.mean(y[indices[i]]))
        consistency = 1.0 - consistency/num_samples

        return consistency

    def _smoothed_base_rates(self, labels, concentration=1.0):
        """Dirichlet-smoothed base rates for each intersecting group in the
        dataset.
        """
        # Dirichlet smoothing parameters
        if concentration < 0:
            raise ValueError("Concentration parameter must be non-negative.")
        num_classes = 2  # binary label dataset
        dirichlet_alpha = concentration / num_classes

        # compute counts for all intersecting groups, e.g. black-women, white-man, etc
        intersect_groups = np.unique(self.dataset.protected_attributes, axis=0)
        num_intersects = len(intersect_groups)
        counts_pos = np.zeros(num_intersects)
        counts_total = np.zeros(num_intersects)
        for i in range(num_intersects):
            condition = [dict(zip(self.dataset.protected_attribute_names,
                                  intersect_groups[i]))]
            counts_total[i] = utils.compute_num_instances(
                    self.dataset.protected_attributes,
                    self.dataset.instance_weights,
                    self.dataset.protected_attribute_names, condition=condition)
            counts_pos[i] = utils.compute_num_pos_neg(
                    self.dataset.protected_attributes, labels,
                    self.dataset.instance_weights,
                    self.dataset.protected_attribute_names,
                    self.dataset.favorable_label, condition=condition)

        # probability of y given S (p(y=1|S))
        return (counts_pos + dirichlet_alpha) / (counts_total + concentration)

    def smoothed_empirical_differential_fairness(self, concentration=1.0):
        """Smoothed EDF from [#foulds18]_.

        Args:
            concentration (float, optional): Concentration parameter for
                Dirichlet smoothing. Must be non-negative.

        Examples:
            To use with non-binary protected attributes, the column must be
            converted to ordinal:

            >>> mapping = {'Black': 0, 'White': 1, 'Asian-Pac-Islander': 2,
            ... 'Amer-Indian-Eskimo': 3, 'Other': 4}
            >>> def map_race(df):
            ...     df['race-num'] = df.race.map(mapping)
            ...     return df
            ...
            >>> adult = AdultDataset(protected_attribute_names=['sex',
            ... 'race-num'], privileged_classes=[['Male'], [1]],
            ... categorical_features=['workclass', 'education',
            ... 'marital-status', 'occupation', 'relationship',
            ... 'native-country', 'race'], custom_preprocessing=map_race)
            >>> metric = BinaryLabelDatasetMetric(adult)
            >>> metric.smoothed_empirical_differential_fairness()
            1.7547611985549287

        References:
            .. [#foulds18] J. R. Foulds, R. Islam, K. N. Keya, and S. Pan,
               "An Intersectional Definition of Fairness," arXiv preprint
               arXiv:1807.08362, 2018.
        """
        sbr = self._smoothed_base_rates(self.dataset.labels, concentration)

        def pos_ratio(i, j):
            return abs(np.log(sbr[i]) - np.log(sbr[j]))

        def neg_ratio(i, j):
            return abs(np.log(1 - sbr[i]) - np.log(1 - sbr[j]))

        # overall DF of the mechanism
        return max(max(pos_ratio(i, j), neg_ratio(i, j))
                   for i in range(len(sbr)) for j in range(len(sbr)) if i != j)

    # ============================== ALIASES ===================================
    def mean_difference(self):
        """Alias of :meth:`statistical_parity_difference`."""
        return self.statistical_parity_difference()


    def rich_subgroup(self, predictions, fairness_def='FP'):
        """Audit dataset with respect to rich subgroups defined by linear thresholds of sensitive attributes

            Args: fairness_def is 'FP' or 'FN' for rich subgroup wrt to false positive or false negative rate.
                  predictions is a hashable tuple of predictions. Typically the labels attribute of a GerryFairClassifier

            Returns: the gamma disparity with respect to the fairness_def.

            Examples: see examples/gerry_plots.ipynb
        """

        auditor = Auditor(self.dataset, fairness_def)

        # make hashable type
        y = array_to_tuple(self.dataset.labels)
        predictions = array_to_tuple(predictions)

        # returns mean(predictions | y = 0) if 'FP' 1-mean(predictions | y = 1) if FN
        metric_baseline = auditor.get_baseline(y, predictions)

        # return the group with the largest disparity
        group = auditor.get_group(predictions, metric_baseline)

        return group.weighted_disparity




from collections import namedtuple
import warnings

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like, is_numeric_dtype


Dataset = namedtuple('Dataset', ['X', 'y'])
WeightedDataset = namedtuple('WeightedDataset', ['X', 'y', 'sample_weight'])

class NumericConversionWarning(UserWarning):
    """Warning used if protected attribute or target is unable to be converted
    automatically to a numeric type."""

def standardize_dataset(df, *, prot_attr, target, sample_weight=None,
        usecols=None, dropcols=None, numeric_only=False, dropna=True):
    """Separate data, targets, and possibly sample weights and populate
    protected attributes as sample properties.

    Args:
        df (pandas.DataFrame): DataFrame with features and, optionally, target.
        prot_attr (label or array-like or list of labels/arrays): Label, array
            of the same length as `df`, or a list containing any combination of
            the two corresponding to protected attribute columns. Even if these
            are dropped from the features, they remain in the index. Column(s)
            indicated by label will be copied from `df`, not dropped. Column(s)
            passed explicitly as arrays will not be added to features.
        target (label or array-like or list of labels/arrays): Label, array of
            the same length as `df`, or a list containing any combination of the
            two corresponding to the target (outcome) variable. Column(s)
            indicated by label will be dropped from features.
        sample_weight (single label or array-like, optional): Name of the column
            containing sample weights or an array of sample weights of the same
            length as `df`. If a label is passed, the column is dropped from
            features. Note: the index of a passed Series will be ignored.
        usecols (list-like, optional): Column(s) to keep. All others are
            dropped.
        dropcols (list-like, optional): Column(s) to drop. Missing labels are
            ignored.
        numeric_only (bool): Drop all non-numeric, non-binary feature columns.
        dropna (bool): Drop rows with NAs.

    Returns:
        collections.namedtuple:

            A tuple-like object where items can be accessed by index or name.
            Contains the following attributes:

            * **X** (`pandas.DataFrame`) -- Feature array.

            * **y** (`pandas.DataFrame` or `pandas.Series`) -- Target array.

            * **sample_weight** (`pandas.Series`, optional) -- Sample weights.

    Note:
        The order of execution for the dropping parameters is: usecols ->
        dropcols -> numeric_only -> dropna.

    Examples:
        >>> import pandas as pd
        >>> from sklearn.linear_model import LinearRegression

        >>> df = pd.DataFrame([[0.5, 1, 1, 0.75], [-0.5, 0, 0, 0.25]],
        ...                   columns=['X', 'y', 'Z', 'w'])
        >>> train = standardize_dataset(df, prot_attr='Z', target='y',
        ...                             sample_weight='w')
        >>> reg = LinearRegression().fit(**train._asdict())

        >>> import numpy as np
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.model_selection import train_test_split
        >>> df = pd.DataFrame(np.hstack(make_classification(n_features=5)))
        >>> X, y = standardize_dataset(df, prot_attr=0, target=5)
        >>> X_tr, X_te, y_tr, y_te = train_test_split(X, y)
    """
    if numeric_only:
        for col in df.select_dtypes('category'):
            if df[col].cat.ordered:
                df[col] = df[col].factorize(sort=True)[0]
                df[col] = df[col].replace(-1, np.nan)

    # protected attribute(s)
    df = df.set_index(prot_attr, drop=False)
    pa = df.index

    # target(s)
    df = df.set_index(target, drop=True)  # utilize set_index logic for mixed types
    y = df.index.to_frame().squeeze()
    df.index = y.index = pa

    # sample weight
    if sample_weight is not None:
        sw = pd.Series(sample_weight) if is_list_like(sample_weight) else \
             df.pop(sample_weight)
        sw.index = pa

    # Column-wise drops
    if usecols:
        if not is_list_like(usecols):
            usecols = [usecols]  # ensure output is DataFrame, not Series
        df = df.loc[:, usecols]
    if dropcols:
        df = df.drop(columns=dropcols, errors='ignore')
    if numeric_only:
        df = df.select_dtypes(['number', 'bool'])
        # warn if nonnumeric prot_attr or target but proceed
        if any(not is_numeric_dtype(dt) for dt in pa.to_frame().dtypes):
            warnings.warn(f"index contains non-numeric:\n{pa.to_frame().dtypes}",
                          category=NumericConversionWarning)
        if any(not is_numeric_dtype(dt) for dt in y.to_frame().dtypes):
            warnings.warn(f"y contains non-numeric column:\n{y.to_frame().dtypes}",
                          category=NumericConversionWarning)

    # Index-wise drops
    if dropna:
        notna = df.notna().all(axis=1) & y.notna() & pa.to_frame().notna().all(axis=1)
        if sample_weight is not None:
            notna &= sw.notna()
            sw = sw.loc[notna]
        df = df.loc[notna]
        y = y.loc[notna]

    for col in df.select_dtypes('category'):
        df[col] = df[col].cat.remove_unused_categories()

    return Dataset(df, y) if sample_weight is None else WeightedDataset(df, y, sw)



import numpy as np

from aif360.algorithms import Transformer
from aif360.metrics import utils


class Reweighing(Transformer):
    """Reweighing is a preprocessing technique that Weights the examples in each
    (group, label) combination differently to ensure fairness before
    classification [4]_.

    References:
        .. [4] F. Kamiran and T. Calders,  "Data Preprocessing Techniques for
           Classification without Discrimination," Knowledge and Information
           Systems, 2012.
    """

    def __init__(self, unprivileged_groups, privileged_groups):
        """
        Args:
            unprivileged_groups (list(dict)): Representation for unprivileged
                group.
            privileged_groups (list(dict)): Representation for privileged group.
        """
        super(Reweighing, self).__init__(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups)

        self.unprivileged_groups = unprivileged_groups
        self.privileged_groups = privileged_groups

        self.w_p_fav = 1.
        self.w_p_unfav = 1.
        self.w_up_fav = 1.
        self.w_up_unfav = 1.

    def fit(self, dataset):
        """Compute the weights for reweighing the dataset.

        Args:
            dataset (BinaryLabelDataset): Dataset containing true labels.

        Returns:
            Reweighing: Returns self.
        """

        (priv_cond, unpriv_cond, fav_cond, unfav_cond,
        cond_p_fav, cond_p_unfav, cond_up_fav, cond_up_unfav) =\
                self._obtain_conditionings(dataset)

        n = np.sum(dataset.instance_weights, dtype=np.float64)
        n_p = np.sum(dataset.instance_weights[priv_cond], dtype=np.float64)
        n_up = np.sum(dataset.instance_weights[unpriv_cond], dtype=np.float64)
        n_fav = np.sum(dataset.instance_weights[fav_cond], dtype=np.float64)
        n_unfav = np.sum(dataset.instance_weights[unfav_cond], dtype=np.float64)

        n_p_fav = np.sum(dataset.instance_weights[cond_p_fav], dtype=np.float64)
        n_p_unfav = np.sum(dataset.instance_weights[cond_p_unfav],
                           dtype=np.float64)
        n_up_fav = np.sum(dataset.instance_weights[cond_up_fav],
                          dtype=np.float64)
        n_up_unfav = np.sum(dataset.instance_weights[cond_up_unfav],
                            dtype=np.float64)

        # reweighing weights
        self.w_p_fav = n_fav*n_p / (n*n_p_fav)
        self.w_p_unfav = n_unfav*n_p / (n*n_p_unfav)
        self.w_up_fav = n_fav*n_up / (n*n_up_fav)
        self.w_up_unfav = n_unfav*n_up / (n*n_up_unfav)

        return self

    def transform(self, dataset):
        """Transform the dataset to a new dataset based on the estimated
        transformation.

        Args:
            dataset (BinaryLabelDataset): Dataset that needs to be transformed.
        Returns:
            dataset (BinaryLabelDataset): Dataset with transformed
                instance_weights attribute.
        """

        dataset_transformed = dataset.copy(deepcopy=True)

        (_, _, _, _, cond_p_fav, cond_p_unfav, cond_up_fav, cond_up_unfav) =\
                            self._obtain_conditionings(dataset)

        # apply reweighing
        dataset_transformed.instance_weights[cond_p_fav] *= self.w_p_fav
        dataset_transformed.instance_weights[cond_p_unfav] *= self.w_p_unfav
        dataset_transformed.instance_weights[cond_up_fav] *= self.w_up_fav
        dataset_transformed.instance_weights[cond_up_unfav] *= self.w_up_unfav

        return dataset_transformed

##############################
#### Supporting functions ####
##############################
    def _obtain_conditionings(self, dataset):
        """Obtain the necessary conditioning boolean vectors to compute
        instance level weights.
        """
        # conditioning
        priv_cond = utils.compute_boolean_conditioning_vector(
                            dataset.protected_attributes,
                            dataset.protected_attribute_names,
                            condition=self.privileged_groups)
        unpriv_cond = utils.compute_boolean_conditioning_vector(
                            dataset.protected_attributes,
                            dataset.protected_attribute_names,
                            condition=self.unprivileged_groups)
        fav_cond = dataset.labels.ravel() == dataset.favorable_label
        unfav_cond = dataset.labels.ravel() == dataset.unfavorable_label

        # combination of label and privileged/unpriv. groups
        cond_p_fav = np.logical_and(fav_cond, priv_cond)
        cond_p_unfav = np.logical_and(unfav_cond, priv_cond)
        cond_up_fav = np.logical_and(fav_cond, unpriv_cond)
        cond_up_unfav = np.logical_and(unfav_cond, unpriv_cond)

        return (priv_cond, unpriv_cond, fav_cond, unfav_cond,
            cond_p_fav, cond_p_unfav, cond_up_fav, cond_up_unfav)



from aif360.explainers import Explainer
from aif360.metrics import Metric


class MetricTextExplainer(Explainer):
    """Class for explaining metric values with text.

    These briefly explain what a metric is and/or how it is calculated unless it
    is obvious (e.g. accuracy) and print the value.

    This class contains text explanations for all metric values regardless of
    which subclass they appear in. This will raise an error if the metric does
    not apply (e.g. calling `true_positive_rate` if
    `type(metric) == DatasetMetric`).
    """

    def __init__(self, metric):
        """Initialize a `MetricExplainer` object.

        Args:
            metric (Metric): The metric to be explained.
        """
        if isinstance(metric, Metric):
            self.metric = metric
        else:
            raise TypeError("metric must be a Metric.")

    def accuracy(self, privileged=None):
        if privileged is None:
            return "Classification accuracy (ACC): {}".format(
                self.metric.accuracy(privileged=privileged))
        return "Classification accuracy on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.accuracy(privileged=privileged))

    def average_abs_odds_difference(self):
        return ("Average absolute odds difference (average of abs(TPR "
                "difference) and abs(FPR difference)): {}".format(
                    self.metric.average_abs_odds_difference()))

    def average_odds_difference(self):
        return ("Average odds difference (average of TPR difference and FPR "
                "difference, 0 = equality of odds): {}".format(
                    self.metric.average_odds_difference()))

    def between_all_groups_coefficient_of_variation(self):
        return "Between-group coefficient of variation: {}".format(
            self.metric.between_all_groups_coefficient_of_variation())

    def between_all_groups_generalized_entropy_index(self, alpha=2):
        return "Between-group generalized entropy index: {}".format(
            self.metric.between_all_groups_generalized_entropy_index(alpha=alpha))

    def between_all_groups_theil_index(self):
        return "Between-group Theil index: {}".format(
            self.metric.between_all_groups_theil_index())

    def between_group_coefficient_of_variation(self):
        return "Between-group coefficient of variation: {}".format(
            self.metric.between_group_coefficient_of_variation())

    def between_group_generalized_entropy_index(self, alpha=2):
        return "Between-group generalized entropy index: {}".format(
            self.metric.between_group_generalized_entropy_index(alpha=alpha))

    def between_group_theil_index(self):
        return "Between-group Theil index: {}".format(
            self.metric.between_group_theil_index())

    def coefficient_of_variation(self):
        return "Coefficient of variation: {}".format(
            self.metric.coefficient_of_variation())

    def consistency(self, n_neighbors=5):
        return "Consistency (Zemel, et al. 2013): {}".format(
            self.metric.consistency(n_neighbors=n_neighbors))

    def disparate_impact(self):
        return ("Disparate impact (probability of favorable outcome for "
                "unprivileged instances / probability of favorable outcome for "
                "privileged instances): {}".format(
                    self.metric.disparate_impact()))

    def error_rate(self, privileged=None):
        if privileged is None:
            return "Error rate (ERR = 1 - ACC): {}".format(
                self.metric.error_rate(privileged=privileged))
        return "Error rate on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.error_rate(privileged))

    def error_rate_difference(self):
        return ("Error rate difference (error rate on unprivileged instances - "
                "error rate on privileged instances): {}".format(
                    self.metric.error_rate_difference()))

    def error_rate_ratio(self):
        return ("Error rate ratio (error rate on unprivileged instances / "
                "error rate on privileged instances): {}".format(
                    self.metric.error_rate_ratio()))

    def false_discovery_rate(self, privileged=None):
        if privileged is None:
            return "False discovery rate (FDR = FP / (FP + TP)): {}".format(
                self.metric.false_discovery_rate(privileged=privileged))
        return "False discovery rate on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.false_discovery_rate(privileged=privileged))

    def false_discovery_rate_difference(self):
        return ("False discovery rate difference (false discovery rate on "
                "unprivileged instances - false discovery rate on privileged "
                "instances): {}".format(
                    self.metric.false_discovery_rate_difference()))

    def false_discovery_rate_ratio(self):
        return ("False discovery rate ratio (false discovery rate on "
                "unprivileged instances - false discovery rate on privileged "
                "instances): {}".format(
                    self.metric.false_discovery_rate_ratio()))

    def false_negative_rate(self, privileged=None):
        if privileged is None:
            return "False negative rate (FNR = FN / (TP + FN)): {}".format(
                self.metric.false_negative_rate(privileged=privileged))
        return "False negative rate on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.false_negative_rate(privileged=privileged))

    def false_negative_rate_difference(self):
        return ("False negative rate difference (false negative rate on "
                "unprivileged instances - false negative rate on privileged "
                "instances): {}".format(
                    self.metric.false_negative_rate_difference()))

    def false_negative_rate_ratio(self):
        return ("False negative rate ratio (false negative rate on "
                "unprivileged instances / false negative rate on privileged "
                "instances): {}".format(
                    self.metric.false_negative_rate_ratio()))

    def false_omission_rate(self, privileged=None):
        if privileged is None:
            return "False omission rate (FOR = FN / (FN + TN)): {}".format(
                self.metric.false_omission_rate(privileged=privileged))
        return "False omission rate on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.false_omission_rate(privileged=privileged))

    def falses_omission_rate_difference(self):
        return ("False omission rate difference (falses omission rate on "
                "unprivileged instances - falses omission rate on privileged "
                "instances): {}".format(
                    self.metric.falses_omission_rate_difference()))

    def false_omission_rate_ratio(self):
        return ("False omission rate ratio (false omission rate on "
                "unprivileged instances - false omission rate on privileged "
                "instances): {}".format(
                    self.metric.false_omission_rate_ratio()))

    def false_positive_rate(self, privileged=None):
        if privileged is None:
            return "False positive rate (FPR = FP / (FP + TN)): {}".format(
                self.metric.false_positive_rate(privileged=privileged))
        return "False positive rate on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.false_positive_rate(privileged=privileged))

    def false_positive_rate_difference(self):
        return ("False positive rate difference (false positive rate on "
                "unprivileged instances - false positive rate on privileged "
                "instances): {}".format(
                    self.metric.false_positive_rate_difference()))

    def false_positive_rate_ratio(self):
        return ("False positive rate ratio (false positive rate on "
                "unprivileged instances / false positive rate on privileged "
                "instances): {}".format(
                    self.metric.false_positive_rate_ratio()))

    def generalized_entropy_index(self, alpha=2):
        return "Generalized entropy index (GE(alpha)): {}".format(
            self.metric.generalized_entropy_index(alpha=alpha))

    def mean_difference(self):
        return ("Mean difference (mean label value on unprivileged instances - "
                "mean label value on privileged instances): {}".format(
                    self.metric.mean_difference()))

    def negative_predictive_value(self, privileged=None):
        if privileged is None:
            return "Negative predictive value (NPV = TN / (TN + FN)): {}".format(
                self.metric.negative_predictive_value(privileged=privileged))
        return "Negative predictive value on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.negative_predictive_value(privileged=privileged))

    def num_false_negatives(self, privileged=None):
        if privileged is None:
            return "Number of false negative instances (FN): {}".format(
                self.metric.num_false_negatives(privileged=privileged))
        return "Number of {} false negative instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_false_negatives(privileged=privileged))

    def num_false_positives(self, privileged=None):
        if privileged is None:
            return "Number of false positive instances (FP): {}".format(
                self.metric.num_false_positives(privileged=privileged))
        return "Number of {} false positive instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_false_positives(privileged=privileged))

    def num_instances(self, privileged=None):
        if privileged is None:
            return "Number of instances: {}".format(
                self.metric.num_instances(privileged=privileged))
        return "Number of {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_instances(privileged=privileged))

    def num_negatives(self, privileged=None):
        if privileged is None:
            return "Number of negative-outcome instances: {}".format(
                self.metric.num_negatives(privileged=privileged))
        return "Number of {} negative-outcome instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_negatives(privileged=privileged))

    def num_positives(self, privileged=None):
        if privileged is None:
            return "Number of positive-outcome instances: {}".format(
                self.metric.num_positives(privileged=privileged))
        return "Number of {} positive-outcome instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_positives(privileged=privileged))

    def num_pred_negatives(self, privileged=None):
        if privileged is None:
            return "Number of negative-outcome instances predicted: {}".format(
                self.metric.num_pred_negatives(privileged=privileged))
        return "Number of {} negative-outcome instances predicted: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_pred_negatives(privileged=privileged))

    def num_pred_positives(self, privileged=None):
        if privileged is None:
            return "Number of positive-outcome instances predicted: {}".format(
                self.metric.num_pred_positives(privileged=privileged))
        return "Number of {} positive-outcome instances predicted: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_pred_positives(privileged=privileged))

    def num_true_negatives(self, privileged=None):
        if privileged is None:
            return "Number of true negative instances (TN): {}".format(
                self.metric.num_true_negatives(privileged=privileged))
        return "Number of {} true negative instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_true_negatives(privileged=privileged))

    def num_true_positives(self, privileged=None):
        if privileged is None:
            return "Number of true positive instances (TP): {}".format(
                self.metric.num_true_positives(privileged=privileged))
        return "Number of {} true positive instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.num_true_positives(privileged=privileged))

    def positive_predictive_value(self, privileged=None):
        if privileged is None:
            return "Positive predictive value (PPV, precision = TP / (TP + FP)): {}".format(
                self.metric.positive_predictive_value(privileged=privileged))
        return "Positive predictive value on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.positive_predictive_value(privileged=privileged))

    def statistical_parity_difference(self):
        return ("Statistical parity difference (probability of favorable "
                "outcome for unprivileged instances - probability of favorable "
                "outcome for privileged instances): {}".format(
                    self.metric.statistical_parity_difference()))

    def theil_index(self):
        return "Theil index (generalized entropy index with alpha = 1): {}".format(
            self.metric.theil_index())

    def true_negative_rate(self, privileged=None):
        if privileged is None:
            return "True negative rate (TNR, specificity = TN / (FP + TN)): {}".format(
                self.metric.true_negative_rate(privileged=privileged))
        return "True negative rate on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.true_negative_rate(privileged=privileged))

    def true_positive_rate(self, privileged=None):
        if privileged is None:
            return "True positive rate (TPR, recall, sensitivity = TP / (TP + FN)): {}".format(
                self.metric.true_positive_rate(privileged=privileged))
        return "True positive rate on {} instances: {}".format(
            'privileged' if privileged else 'unprivileged',
            self.metric.true_positive_rate(privileged=privileged))

    def true_positive_rate_difference(self):
        return ("True positive rate difference (true positive rate on "
                "unprivileged instances - true positive rate on privileged "
                "instances): {}".format(
                    self.metric.true_positive_rate_difference()))

    # ============================== ALIASES ===================================
    def equal_opportunity_difference(self):
        return self.true_positive_rate_difference()

    def power(self, privileged=None):
        return self.num_true_positives(privileged=privileged)

    def precision(self, privileged=None):
        return self.positive_predictive_value(privileged=privileged)

    def recall(self, privileged=None):
        return self.true_positive_rate(privileged=privileged)

    def sensitivity(self, privileged=None):
        return self.true_positive_rate(privileged=privileged)

    def specificity(self, privileged=None):
        return self.true_negative_rate(privileged=privileged)
